module params
  implicit none

  !  Useful constants
  integer, parameter :: &
       fdp=kind(0.0d0), variablelength=256, errorlength=256

  real(fdp), parameter :: pi = 3.141592653589793238d0, pi2 = pi/2.0d0, &
       pi4 = pi/4.0d0, eps = 1.0d-16, inf = huge(inf), &
       el=0.5772156649015329d0, sqrt2 = 1.4142135623730950488d0, zero=tiny(zero)


  character(len=*), parameter :: &
       operatorchars='+-/*^><=!', numberchars='1234567890.', &
       exponentchars='ed', lowerchars='abcdefghijklmnopqrstuvwxyz', &
       upperchars='ABCDEFGHIJKLMNOPQRSTUVWXYZ', bracketchars='()[]{}', &
       specialchars=',_', funcvarchars=lowerchars//upperchars, &
       allowedchars=operatorchars//numberchars//funcvarchars//bracketchars//specialchars

  ! Controlling definitions
  integer, parameter :: &
       number_=10, bracket_=11, nop_=12, comma_=13, end_=14

  !  Bracket definitions
  integer, parameter :: &
       bracketstart=20, &
       roundbracket_=20, squarebracket_=21, squigbracket_=22, &
       bracketend=22

  !  Expression definitions
  integer, parameter :: &
       expressionlength=8, &
       expressionstart=100, &
       anint_=100, log10_=101, sqrt_=102, acos_=103, asin_=104, &
       atan_=105, cosh_=106, sinh_=107, tanh_=108, aint_=109, &
       cos_=110, sin_=111, tan_=112, exp_=113, log_=114, abs_=115, &
       delta_=116, step_=117, hat_=118, max_=119, min_=120, &
       besj0_=121, besj1_=122, besjn_=123, besy0_=124, besy1_=125, &
       besyn_=126, logn_=127, erf_=128, erfc_=129, lgamma_=130, &
       gamma_=131, csch_=132, sech_=133, coth_=134, if_=135, &
       gauss_=136, sinc_=137, besi0_=138, besi1_=139, besk0_=140, &
       besk1_=141, ierfc_=142, besin_=143, beskn_=144, cbrt_=145, &
       fresc_=146, fress_=147, expi_=148, sini_=149, cosi_=150, &
       logi_=151, ierf_=152, elle_=153, ellk_=154, ielle_=155, iellf_=156, &
       expressionend=156

  character(len=expressionlength), dimension(expressionstart:expressionend), parameter :: &
       expressions=(/ 'anint(  ', 'log10(  ', 'sqrt(   ', 'acos(   ', &
       'asin(   ', 'atan(   ', 'cosh(   ', 'sinh(   ', 'tanh(   ', 'aint(   ', &
       'cos(    ', 'sin(    ', 'tan(    ', 'exp(    ', 'log(    ', 'abs(    ', &
       'delta(  ', 'step(   ', 'hat(    ', 'max(    ', 'min(    ', 'besj0(  ', &
       'besj1(  ', 'besjn(  ', 'besy0(  ', 'besy1(  ', 'besyn(  ', 'logn(   ', &
       'erf(    ', 'erfc(   ', 'lgamma( ', 'gamma(  ', 'csch(   ', 'sech(   ', &
       'coth(   ', 'if(     ', 'gauss(  ', 'sinc(   ', 'besi0(  ', 'besi1(  ', &
       'besk0(  ', 'besk1(  ', 'ierfc(  ', 'besin(  ', 'beskn(  ', 'cbrt(   ', &
       'fresc(  ', 'fress(  ', 'expi(   ', 'sini(   ', 'cosi(   ', 'logi(   ', &
       'ierf(   ', 'elle(   ', 'ellk(   ', 'ielle(  ', 'iellf(  ' &
       /)
      

  integer, dimension(expressionstart:expressionend), parameter :: &
       expressionslen=len_trim(expressions), &
       commandcomps=(/ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, &
       1, 1, 1, 1, 1, 2, 2, 1, 1, 2, 1, 1, 2, 2, 1, 1, 1, 1, 1, 1, &
       1, 3, 1, 1, 1, 1, 1, 1, 1, 2, 2, 1, 1, 1, 1, 1, 1, 1, &
       1, 1, 1, 2, 2 /)
  integer, parameter :: &
       maxcommandcomps=4

  !  Operators definitions
  integer, parameter :: &
       operatorlength=2, &
       operatorstart=500, &
       conditionalstart=500, &
       neq_=500, eqeq_=501, gteq_=502, eqgt_=503, lteq_=504, eqlt_=505, eq_=506, gt_=507, lt_=508,&
       conditionalend=508, &
       opsstart=509, &
       pow_=509, caret_=510, mult_=511, div_=512, &
       opsend=512, &
       pmstart=513, &
       plus_=513, minus_=514, &
       pmend=514, &
       operatorend=514

  character(len=operatorlength), dimension(operatorstart:operatorend), parameter :: &
       operators=(/ '!=', '==', '>=', '=>', '<=', '=<', '= ', &
       '> ', '< ', '**', '^ ', '* ', '/ ', '+ ', '- ' /)

  integer, dimension(operatorstart:operatorend), parameter :: &
       operatorsorder=(/ 10, 10, 10, 10, 10, 10, 10, 10, 10, &
       10000, 10000, 1000, 1000, 100, 100 /), &
       operatorslen=len_trim(operators)

  !  Variable definitions
  integer, parameter :: &
       variablestart=1000, &
       variableend=9999

  ! Separates logical blocks of code
  integer, dimension(10), parameter :: blockseparator = (/ comma_, neq_, eqeq_, gteq_, &
       eqgt_, lteq_, eqlt_, eq_, gt_, lt_ /) 

end module params

module extrafunc
  !
  !  Extra functions, they must take real input and return real
  !

  !
  !  Many of these functions were taken from Takuya OOURA (email: ooura@mmm.t.u-tokyo.ac.jp)
  !  with modifications http://momonga.t.u-tokyo.ac.jp/~ooura/index.html
  !
  !  Some of these functions were programmed with ideas from Numerical Recipes
  !
  !  Other functions came from Shanjie Zhang and Jianming Jin and their specfun 
  !  functions http://jin.ece.uiuc.edu/routines/routines.html with modifications
  !

  use params

  implicit none
  
  private
  public :: f_besj0, f_besj1, f_besjn, f_besy0, f_besy1, f_besyn,&
       & f_erf, f_gamma, f_lgamma, f_cbrt, f_besi0, f_besi1, f_besin,&
       & f_besk0, f_besk1, f_beskn, f_ierfc, f_fresc, f_fress, &
       & f_expi, f_sini, f_cosi, f_logi, f_ierf, f_elle, f_ellk, f_ielle, f_iellf

contains

  recursive function f_besj0(x) result(f)
    !
    ! Bessel function 1st kind, order 0
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    integer :: k
    real(fdp) :: t, w, v, y, theta
    real(fdp), dimension(0:7), parameter :: a = (/ &
         &-0.0000000000023655394d0, 0.0000000004708898680d0, &
         &-0.0000000678167892231d0, 0.0000067816840038636d0, &
         &-0.0004340277777716935d0, 0.0156249999999992397d0, &
         &-0.2499999999999999638d0, 0.9999999999999999997d0 /)
    real(fdp), dimension(0:64), parameter :: b = (/&
         & 0.0000000000626681117d0, -0.0000000022270614428d0,&
         & 0.0000000662981656302d0, -0.0000016268486502196d0,&
         & 0.0000321978384111685d0, -0.0005005237733315830d0,&
         & 0.0059060313537449816d0, -0.0505265323740109701d0,&
         & 0.2936432097610503985d0, -1.0482565081091638637d0,&
         & 1.9181123286040428113d0, -1.1319199475221700100d0, &
         &-0.1965480952704682000d0, 0.0000000000457457332d0, &
         &-0.0000000015814772025d0, 0.0000000455487446311d0, &
         &-0.0000010735201286233d0, 0.0000202015179970014d0, &
         &-0.0002942392368203808d0, 0.0031801987726150648d0, &
         &-0.0239875209742846362d0, 0.1141447698973777641d0, &
         &-0.2766726722823530233d0, 0.1088620480970941648d0,&
         & 0.5136514645381999197d0, -0.2100594022073706033d0,&
         & 0.0000000000331366618d0, -0.0000000011119090229d0,&
         & 0.0000000308823040363d0, -0.0000006956602653104d0,&
         & 0.0000123499947481762d0, -0.0001662951945396180d0,&
         & 0.0016048663165678412d0, -0.0100785479932760966d0,&
         & 0.0328996815223415274d0, -0.0056168761733860688d0, &
         &-0.2341096400274429386d0, 0.2551729256776404262d0,&
         & 0.2288438186148935667d0, 0.0000000000238007203d0, &
         &-0.0000000007731046439d0, 0.0000000206237001152d0, &
         &-0.0000004412291442285d0, 0.0000073107766249655d0, &
         &-0.0000891749801028666d0, 0.0007341654513841350d0, &
         &-0.0033303085445352071d0, 0.0015425853045205717d0,&
         & 0.0521100583113136379d0, -0.1334447768979217815d0, &
         &-0.1401330292364750968d0, 0.2685616168804818919d0,&
         & 0.0000000000169355950d0, -0.0000000005308092192d0,&
         & 0.0000000135323005576d0, -0.0000002726650587978d0,&
         & 0.0000041513240141760d0, -0.0000443353052220157d0,&
         & 0.0002815740758993879d0, -0.0004393235121629007d0, &
         &-0.0067573531105799347d0, 0.0369141914660130814d0,&
         & 0.0081673361942996237d0, -0.2573381285898881860d0,&
         & 0.0459580257102978932d0 /) 
    real(fdp), dimension(0:69), parameter :: c = (/ &
         &-0.00000000003009451757d0, -0.00000000014958003844d0,&
         & 0.00000000506854544776d0, 0.00000001863564222012d0, &
         &-0.00000060304249068078d0, -0.00000147686259937403d0,&
         & 0.00004714331342682714d0, 0.00006286305481740818d0, &
         &-0.00214137170594124344d0, -0.00089157336676889788d0,&
         & 0.04508258728666024989d0, -0.00490362805828762224d0, &
         &-0.27312196367405374426d0, 0.04193925184293450356d0, &
         &-0.00000000000712453560d0, -0.00000000041170814825d0,&
         & 0.00000000138012624364d0, 0.00000005704447670683d0, &
         &-0.00000019026363528842d0, -0.00000533925032409729d0,&
         & 0.00001736064885538091d0, 0.00030692619152608375d0, &
         &-0.00092598938200644367d0, -0.00917934265960017663d0,&
         & 0.02287952522866389076d0, 0.10545197546252853195d0, &
         &-0.16126443075752985095d0, -0.19392874768742235538d0,&
         & 0.00000000002128344556d0, -0.00000000031053910272d0, &
         &-0.00000000334979293158d0, 0.00000004507232895050d0,&
         & 0.00000036437959146427d0, -0.00000446421436266678d0, &
         &-0.00002523429344576552d0, 0.00027519882931758163d0,&
         & 0.00097185076358599358d0, -0.00898326746345390692d0, &
         &-0.01665959196063987584d0, 0.11456933464891967814d0,&
         & 0.07885001422733148815d0, -0.23664819446234712621d0,&
         & 0.00000000003035295055d0, 0.00000000005486066835d0, &
         &-0.00000000501026824811d0, -0.00000000501246847860d0,&
         & 0.00000058012340163034d0, 0.00000016788922416169d0, &
         &-0.00004373270270147275d0, 0.00001183898532719802d0,&
         & 0.00189863342862291449d0, -0.00113759249561636130d0, &
         &-0.03846797195329871681d0, 0.02389746880951420335d0,&
         & 0.22837862066532347461d0, -0.06765394811166522844d0,&
         & 0.00000000001279875977d0, 0.00000000035925958103d0, &
         &-0.00000000228037105967d0, -0.00000004852770517176d0,&
         & 0.00000028696428000189d0, 0.00000440131125178642d0, &
         &-0.00002366617753349105d0, -0.00024412456252884129d0,&
         & 0.00113028178539430542d0, 0.00708470513919789080d0, &
         &-0.02526914792327618386d0, -0.08006137953480093426d0,&
         & 0.16548380461475971846d0, 0.14688405470042110229d0 /)
    real(fdp), dimension(0:51), parameter :: d = (/&
         & 1.059601355592185731d-14, -2.71150591218550377d-13,&
         & 8.6514809056201638d-12, -4.6264028554286627d-10,&
         & 5.0815403835647104d-8, -1.76722552048141208d-5,&
         & 0.16286750396763997378d0, 2.949651820598278873d-13, &
         &-8.818215611676125741d-12, 3.571119876162253451d-10, &
         &-2.631924120993717060d-8, 4.709502795656698909d-6, &
         &-5.208333333333283282d-3, 7.18344107717531977d-15, &
         &-2.51623725588410308d-13, 8.6017784918920604d-12, &
         &-4.6256876614290359d-10, 5.0815343220437937d-8, &
         &-1.76722551764941970d-5, 0.16286750396763433767d0,&
         & 2.2327570859680094777d-13, -8.464594853517051292d-12,&
         & 3.563766464349055183d-10, -2.631843986737892965d-8,&
         & 4.709502342288659410d-6, -5.2083333332278466225d-3,&
         & 5.15413392842889366d-15, -2.27740238380640162d-13,&
         & 8.4827767197609014d-12, -4.6224753682737618d-10,&
         & 5.0814848128929134d-8, -1.76722547638767480d-5,&
         & 0.16286750396748926663d0, 1.7316195320192170887d-13, &
         &-7.971122772293919646d-12, 3.544039469911895749d-10, &
         &-2.631443902081701081d-8, 4.709498228695400603d-6, &
         &-5.2083333315143653610d-3, 3.84653681453798517d-15, &
         &-2.04464520778789011d-13, 8.3089298605177838d-12, &
         &-4.6155016158412096d-10, 5.0813263696466650d-8, &
         &-1.76722528311426167d-5, 0.16286750396650065930d0,&
         & 1.3797879972460878797d-13, -7.448089381011684812d-12,&
         & 3.512733797106959780d-10, -2.630500895563592722d-8,&
         & 4.709483934775839193d-6, -5.2083333227940760113d-3 /)

    w = abs(x)

    if (w  <  1) then
       t = w * w
       y = ((((((a(0) * t + a(1)) * t + a(2)) * t + a(3)) * t +&
            & a(4)) * t + a(5)) * t + a(6)) * t + a(7)
    else if (w  <  8.5d0) then
       t = w * w * 0.0625d0
       k = int(t)
       t = t - (k + 0.5d0)
       k = k * 13
       y = (((((((((((b(k) * t + b(k + 1)) * t + b(k + 2)) * t +&
            & b(k + 3)) * t + b(k + 4)) * t + b(k + 5)) * t + b(k&
            & + 6)) * t + b(k + 7)) * t + b(k + 8)) * t + b(k +&
            & 9)) * t + b(k + 10)) * t + b(k + 11)) * t + b(k +&
            & 12)
    else if (w  <  12.5d0) then
       k = int(w)
       t = w - (k + 0.5d0)
       k = 14 * (k - 8)
       y = ((((((((((((c(k) * t + c(k + 1)) * t + c(k + 2)) * t +&
            & c(k + 3)) * t + c(k + 4)) * t + c(k + 5)) * t + c(k&
            & + 6)) * t + c(k + 7)) * t + c(k + 8)) * t + c(k +&
            & 9)) * t + c(k + 10)) * t + c(k + 11)) * t + c(k +&
            & 12)) * t + c(k + 13)
    else
       v = 24 / w
       t = v * v
       k = 13 * (int(t))
       y = ((((((d(k) * t + d(k + 1)) * t + d(k + 2)) * t + d(k +&
            & 3)) * t + d(k + 4)) * t + d(k + 5)) * t + d(k + 6))&
            & * sqrt(v)
       theta = (((((d(k + 7) * t + d(k + 8)) * t + d(k + 9)) * t &
            &+ d(k + 10)) * t + d(k + 11)) * t + d(k + 12)) * v -&
            & pi4
       y = y * cos(w + theta)
    end if

    f = y
  end function f_besj0



  recursive function f_besj1(x) result(f)
    !
    ! Bessel function 1st kind, order 1
    !
    real(fdp), intent(in) :: x
    real(fdp) :: f
    
    integer :: k
    real(fdp) :: t, w, v, y, theta
    real(fdp), dimension(0:7), parameter :: a = (/ &
         &-0.00000000000014810349d0, 0.00000000003363594618d0, &
         &-0.00000000565140051697d0, 0.00000067816840144764d0, &
         &-0.00005425347222188379d0, 0.00260416666666662438d0, &
         &-0.06249999999999999799d0, 0.49999999999999999998d0 /)
    real(fdp), dimension(0:64), parameter :: b = (/&
         & 0.00000000000243721316d0, -0.00000000009400554763d0,&
         & 0.00000000306053389980d0, -0.00000008287270492518d0,&
         & 0.00000183020515991344d0, -0.00003219783841164382d0,&
         & 0.00043795830161515318d0, -0.00442952351530868999d0,&
         & 0.03157908273375945955d0, -0.14682160488052520107d0,&
         & 0.39309619054093640008d0, -0.47952808215101070280d0,&
         & 0.14148999344027125140d0, 0.00000000000182119257d0, &
         &-0.00000000006862117678d0, 0.00000000217327908360d0, &
         &-0.00000005693592917820d0, 0.00000120771046483277d0, &
         &-0.00002020151799736374d0, 0.00025745933218048448d0, &
         &-0.00238514907946126334d0, 0.01499220060892984289d0, &
         &-0.05707238494868888345d0, 0.10375225210588234727d0, &
         &-0.02721551202427354117d0, -0.06420643306727498985d0,&
         & 0.000000000001352611196d0, -0.000000000049706947875d0,&
         & 0.000000001527944986332d0, -0.000000038602878823401d0,&
         & 0.000000782618036237845d0, -0.000012349994748451100d0,&
         & 0.000145508295194426686d0, -0.001203649737425854162d0,&
         & 0.006299092495799005109d0, -0.016449840761170764763d0,&
         & 0.002106328565019748701d0, 0.058527410006860734650d0, &
         &-0.031896615709705053191d0, 0.000000000000997982124d0, &
         &-0.000000000035702556073d0, 0.000000001062332772617d0, &
         &-0.000000025779624221725d0, 0.000000496382962683556d0, &
         &-0.000007310776625173004d0, 0.000078028107569541842d0, &
         &-0.000550624088538081113d0, 0.002081442840335570371d0, &
         &-0.000771292652260286633d0, -0.019541271866742634199d0,&
         & 0.033361194224480445382d0, 0.017516628654559387164d0,&
         & 0.000000000000731050661d0, -0.000000000025404499912d0,&
         & 0.000000000729360079088d0, -0.000000016915375004937d0,&
         & 0.000000306748319652546d0, -0.000004151324014331739d0,&
         & 0.000038793392054271497d0, -0.000211180556924525773d0,&
         & 0.000274577195102593786d0, 0.003378676555289966782d0, &
         &-0.013842821799754920148d0, -0.002041834048574905921d0,&
         & 0.032167266073736023299d0 /)
    real(fdp), dimension(0:69), parameter :: c = (/ &
         &-0.00000000001185964494d0, 0.00000000039110295657d0,&
         & 0.00000000180385519493d0, -0.00000005575391345723d0, &
         &-0.00000018635897017174d0, 0.00000542738239401869d0,&
         & 0.00001181490114244279d0, -0.00033000319398521070d0, &
         &-0.00037717832892725053d0, 0.01070685852970608288d0,&
         & 0.00356629346707622489d0, -0.13524776185998074716d0,&
         & 0.00980725611657523952d0, 0.27312196367405374425d0, &
         &-0.00000000003029591097d0, 0.00000000009259293559d0,&
         & 0.00000000496321971223d0, -0.00000001518137078639d0, &
         &-0.00000057045127595547d0, 0.00000171237271302072d0,&
         & 0.00004271400348035384d0, -0.00012152454198713258d0, &
         &-0.00184155714921474963d0, 0.00462994691003219055d0,&
         & 0.03671737063840232452d0, -0.06863857568599167175d0, &
         &-0.21090395092505707655d0, 0.16126443075752985095d0, &
         &-0.00000000002197602080d0, -0.00000000027659100729d0,&
         & 0.00000000374295124827d0, 0.00000003684765777023d0, &
         &-0.00000045072801091574d0, -0.00000327941630669276d0,&
         & 0.00003571371554516300d0, 0.00017664005411843533d0, &
         &-0.00165119297594774104d0, -0.00485925381792986774d0,&
         & 0.03593306985381680131d0, 0.04997877588191962563d0, &
         &-0.22913866929783936544d0, -0.07885001422733148814d0,&
         & 0.00000000000516292316d0, -0.00000000039445956763d0, &
         &-0.00000000066220021263d0, 0.00000005511286218639d0,&
         & 0.00000005012579400780d0, -0.00000522111059203425d0, &
         &-0.00000134311394455105d0, 0.00030612891890766805d0, &
         &-0.00007103391195326182d0, -0.00949316714311443491d0,&
         & 0.00455036998246516948d0, 0.11540391585989614784d0, &
         &-0.04779493761902840455d0, -0.22837862066532347460d0,&
         & 0.00000000002697817493d0, -0.00000000016633326949d0, &
         &-0.00000000433134860350d0, 0.00000002508404686362d0,&
         & 0.00000048528284780984d0, -0.00000258267851112118d0, &
         &-0.00003521049080466759d0, 0.00016566324273339952d0,&
         & 0.00146474737522491617d0, -0.00565140892697147306d0, &
         &-0.02833882055679300400d0, 0.07580744376982855057d0,&
         & 0.16012275906960187978d0, -0.16548380461475971845d0 /)
    real(fdp), dimension(0:51), parameter :: d = (/ &
         &-1.272346002224188092d-14, 3.370464692346669075d-13, &
         &-1.144940314335484869d-11, 6.863141561083429745d-10, &
         &-9.491933932960924159d-8, 5.301676561445687562d-5,&
         & 0.1628675039676399740d0, -3.652982212914147794d-13,&
         & 1.151126750560028914d-11, -5.165585095674343486d-10,&
         & 4.657991250060549892d-8, -1.186794704692706504d-5,&
         & 1.562499999999994026d-2, -8.713069680903981555d-15,&
         & 3.140780373478474935d-13, -1.139089186076256597d-11,&
         & 6.862299023338785566d-10, -9.491926788274594674d-8,&
         & 5.301676558106268323d-5, 0.1628675039676466220d0, &
         &-2.792555727162752006d-13, 1.108650207651756807d-11, &
         &-5.156745588549830981d-10, 4.657894859077370979d-8, &
         &-1.186794650130550256d-5, 1.562499999987299901d-2, &
         &-6.304859171204770696d-15, 2.857249044208791652d-13, &
         &-1.124956921556753188d-11, 6.858482894906716661d-10, &
         &-9.491867953516898460d-8, 5.301676509057781574d-5,&
         & 0.1628675039678191167d0, -2.185193490132496053d-13,&
         & 1.048820673697426074d-11, -5.132819367467680132d-10,&
         & 4.657409437372994220d-8, -1.186794150862988921d-5,&
         & 1.562499999779270706d-2, -4.740417209792009850d-15,&
         & 2.578715253644144182d-13, -1.104148898414138857d-11,&
         & 6.850134201626289183d-10, -9.491678234174919640d-8,&
         & 5.301676277588728159d-5, 0.1628675039690033136d0, &
         &-1.755122057493842290d-13, 9.848723331445182397d-12, &
         &-5.094535425482245697d-10, 4.656255982268609304d-8, &
         &-1.186792402114394891d-5, 1.562499998712198636d-2 /)

    w = abs(x)

    if (w  <  1) then
       t = w * w
       y = (((((((a(0) * t + a(1)) * t + a(2)) * t + a(3)) * t +&
            & a(4)) * t + a(5)) * t + a(6)) * t + a(7)) * w
    else if (w  <  8.5d0) then
       t = w * w * 0.0625d0
       k = int(t)
       t = t - (k + 0.5d0)
       k = k * 13
       y = ((((((((((((b(k) * t + b(k + 1)) * t + b(k + 2)) * t +&
            & b(k + 3)) * t + b(k + 4)) * t + b(k + 5)) * t + b(k&
            & + 6)) * t + b(k + 7)) * t + b(k + 8)) * t + b(k +&
            & 9)) * t + b(k + 10)) * t + b(k + 11)) * t + b(k +&
            & 12)) * w
    else if (w  <  12.5d0) then
       k = int(w)
       t = w - (k + 0.5d0)
       k = 14 * (k - 8)
       y = ((((((((((((c(k) * t + c(k + 1)) * t + c(k + 2)) * t +&
            & c(k + 3)) * t + c(k + 4)) * t + c(k + 5)) * t + c(k&
            & + 6)) * t + c(k + 7)) * t + c(k + 8)) * t + c(k +&
            & 9)) * t + c(k + 10)) * t + c(k + 11)) * t + c(k +&
            & 12)) * t + c(k + 13)
    else
       v = 24 / w
       t = v * v
       k = 13 * (int(t))
       y = ((((((d(k) * t + d(k + 1)) * t + d(k + 2)) * t + d(k +&
            & 3)) * t + d(k + 4)) * t + d(k + 5)) * t + d(k + 6))&
            & * sqrt(v)
       theta = (((((d(k + 7) * t + d(k + 8)) * t + d(k + 9)) * t &
            &+ d(k + 10)) * t + d(k + 11)) * t + d(k + 12)) * v -&
            & pi4
       y = y * sin(w + theta)
    end if

    if (x  <  0) y = -y

    f = y
  end function f_besj1



  recursive function f_besjn(n, x) result(f)
    !
    ! Bessel function 1st kind, order n
    !

    real(fdp), intent(in) :: n, x
    real(fdp) :: f

    integer, parameter :: iexp=maxexponent(x)/2
    real(fdp), parameter :: iacc = 50.0d0
    integer :: j, jsum, m, nn
    real(fdp) :: ax, bj, bjm, bjp, summ, tox
    
    nn = nint(n)

    if (nn == 0) then
       f = f_besj0(x)
    else if (nn == 1) then
       f = f_besj1(x)
    else
       ax=abs(x)
       if (ax*ax <= 8.0d0*tiny(x)) then
          f = 0.0d0
       else if (ax > n) then
          tox = 2.0d0/ax
          bjm = f_besj0(ax)
          bj = f_besj1(ax)
          do j = 1, nn-1
             bjp = j*tox*bj-bjm
             bjm = bj
             bj = bjp
          end do
          f = bj
       else
          tox = 2.0d0/ax
          m = 2*((nn+int(sqrt(iacc*n)))/2)
          f = 0.0
          jsum = 0
          summ = 0.0
          bjp = 0.0
          bj = 1.0
          do j = m, 1, -1
             bjm = j*tox*bj-bjp
             bjp = bj
             bj = bjm
             if (exponent(bj) > iexp) then
                bj = scale(bj, -iexp)
                bjp = scale(bjp, -iexp)
                f = scale(f, -iexp)
                summ = scale(summ, -iexp)
             end if
             if (jsum /= 0) summ = summ+bj
             jsum = 1-jsum
             if (j == nn) f = bjp
          end do
          summ = 2.0d0*summ-bj
          f = f/summ
       end if
       f = merge(-f, f, x < 0.0d0 .and. mod(nn,2) == 1)
    end if
  end function f_besjn
  


  recursive function f_besy0(x) result(f)
    !
    ! Bessel function 2nd kind, order 0
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f
    
    real(fdp) :: w, t, v, y, theta
    integer :: k
    real(fdp), dimension(0:15), parameter :: a = (/ &
         &-0.00000000000151249795d0, 0.00000000029979612902d0, &
         &-0.00000004317352912436d0, 0.00000431735413787068d0, &
         &-0.00027631066508933090d0, 0.00994718394324338940d0, &
         &-0.15915494309189533339d0, 0.63661977236758134306d0,&
         & 0.00000000000409490035d0, -0.00000000076925095943d0,&
         & 0.00000010358472550303d0, -0.00000949500519343105d0,&
         & 0.00053860266685948738d0, -0.01607396802593822992d0,&
         & 0.17760601686906713536d0, -0.07380429510868722524d0 /)
    real(fdp), dimension(0:111), parameter :: b = (/ &
         &-0.00000000002958527319d0, -0.00000000004799424787d0,&
         & 0.00000000070761365903d0, 0.00000000786916310954d0,&
         & 0.00000004406321746330d0, 0.00000010474207054310d0, &
         &-0.00000070359581447993d0, -0.00001068225180236166d0, &
         &-0.00007636033201484949d0, -0.00034318294473105478d0, &
         &-0.00050019381740765567d0, 0.00898034680285547482d0,&
         & 0.14783488851798565262d0, 0.01218096458136948754d0,  &
         &-0.00000000003790394288d0, -0.00000000083012324083d0, &
         &-0.00000000416344143065d0, -0.00000000159750317942d0,&
         & 0.00000011897479804282d0, 0.00000089462034665663d0,&
         & 0.00000279277855080310d0, -0.00000710267818373816d0, &
         &-0.00014158631248747538d0, -0.00089512296835766028d0, &
         &-0.00286419212921734238d0, 0.00448696141199596815d0,&
         & 0.16247276853497028694d0, 0.16806523426268420517d0,&
         & 0.00000000049508820408d0, 0.00000000240256475451d0, &
         &-0.00000000341287024760d0, -0.00000007926555593644d0, &
         &-0.00000027510687602104d0, 0.00000082850014912691d0,&
         & 0.00001202838962100801d0, 0.00004446440998573947d0, &
         &-0.00006345799532264893d0, -0.00153636349657759957d0, &
         &-0.00784582749467635199d0, -0.01091113270659651752d0,&
         & 0.15855078058220179592d0, 0.33112075987570150227d0, &
         &-0.00000000121381163034d0, -0.00000000185252445637d0,&
         & 0.00000002782181715752d0, 0.00000009983716245339d0, &
         &-0.00000051382328105461d0, -0.00000413045112612025d0,&
         & 0.00000069219807212556d0, 0.00011346881914173059d0,&
         & 0.00044857142622311310d0, -0.00075862241142011713d0, &
         &-0.01330170710405210993d0, -0.04337207915110066583d0,&
         & 0.10708263075506099964d0, 0.46937172516119821029d0,&
         & 0.00000000123954313286d0, -0.00000000131506473470d0, &
         &-0.00000003208158247443d0, 0.00000006936208476862d0,&
         & 0.00000096326321605564d0, -0.00000186813677726279d0, &
         &-0.00003296596487362508d0, -0.00001027532398838717d0,&
         & 0.00088296805857821394d0, 0.00288502916411875494d0, &
         &-0.00981551522729451135d0, -0.08175499367255420548d0, &
         &-0.01970902432425660780d0, 0.51958016003260732459d0, &
         &-0.00000000050293922567d0, 0.00000000259058191150d0,&
         & 0.00000000380069015768d0, -0.00000012579287605830d0,&
         & 0.00000029666593806065d0, 0.00000532199410651341d0, &
         &-0.00001466013674557952d0, -0.00021196179800800308d0,&
         & 0.00014889075157975250d0, 0.00598670133725335804d0,&
         & 0.00917434640154761191d0, -0.08592436550224288681d0, &
         &-0.19700894914535846940d0, 0.41202451505149651586d0,&
         & 0.00000000002689416111d0, -0.00000000060225798842d0,&
         & 0.00000000716632287642d0, -0.00000003038834206093d0, &
         &-0.00000049829955736697d0, 0.00000364678661079850d0,&
         & 0.00002599276311771812d0, -0.00016374249229498868d0, &
         &-0.00112385259950125883d0, 0.00342431039748960465d0,&
         & 0.03016663018029814938d0, -0.02432512430353342927d0, &
         &-0.31797371916576712163d0, 0.14418001571733803493d0,&
         & 0.00000000000194632210d0, -0.00000000013451327719d0,&
         & 0.00000000336581427068d0, 0.00000002163920016772d0, &
         &-0.00000050868502638387d0, -0.00000126562372408989d0,&
         & 0.00003554054033687039d0, 0.00007490048474688881d0, &
         &-0.00142380042008861330d0, -0.00355314274445060456d0,&
         & 0.03042001885994619102d0, 0.07365497405664882878d0, &
         &-0.26882214651298246524d0, -0.16578636480856581212d0 /)
    real(fdp), dimension(0:125), parameter :: c = (/&
         & 0.00000000014133441141d0, -0.00000000116239182574d0,&
         & 0.00000000560813675548d0, 0.00000004361111919906d0, &
         &-0.00000019980743680592d0, -0.00000623828267379045d0,&
         & 0.00002659779111113435d0, 0.00034429768949740887d0, &
         &-0.00128058451746064875d0, -0.01165906901727504569d0,&
         & 0.03800023545011044967d0, 0.13079665132249175616d0, &
         &-0.30099732306965462304d0, -0.19470500862950453349d0,&
         & 0.00000000002985070388d0, -0.00000000043935370028d0, &
         &-0.00000000270094988368d0, 0.00000005270139367180d0,&
         & 0.00000035596645009533d0, -0.00000559393055101660d0, &
         &-0.00002415386269153308d0, 0.00034989509210648774d0,&
         & 0.00098204687587121096d0, -0.01241793885175385487d0, &
         &-0.01398519840974389960d0, 0.16758045653582919005d0,&
         & 0.02375823895638961834d0, -0.33948059288191103829d0,&
         & 0.00000000003440543098d0, -0.00000000001492311224d0, &
         &-0.00000000550307983816d0, 0.00000000288861169685d0,&
         & 0.00000065975576340131d0, -0.00000064253433004282d0, &
         &-0.00005095354686815779d0, 0.00006355454993858760d0,&
         & 0.00231750171436134067d0, -0.00344140664586595266d0, &
         &-0.04796153689875432703d0, 0.06553727330877767204d0,&
         & 0.27409127395927545297d0, -0.17324243491898233567d0,&
         & 0.00000000001869953108d0, 0.00000000035678205287d0, &
         &-0.00000000323369443017d0, -0.00000004932837601739d0,&
         & 0.00000040636932259569d0, 0.00000455294558480472d0, &
         &-0.00003376596464782138d0, -0.00025758322585307069d0,&
         & 0.00167416927479443727d0, 0.00735300543429220468d0, &
         &-0.03904554680817849396d0, -0.07593187710651205994d0,&
         & 0.25912851048611625179d0, 0.11731328614820863082d0, &
         &-0.00000000001168499781d0, 0.00000000040491570620d0,&
         & 0.00000000176042185153d0, -0.00000005796810963324d0, &
         &-0.00000017664264701199d0, 0.00000565232278687159d0,&
         & 0.00001060939366068237d0, -0.00034382320897779425d0, &
         &-0.00028786871040915476d0, 0.01103696602501040460d0,&
         & 0.00105742483935856071d0, -0.13664188676516064192d0,&
         & 0.02616867939853747003d0, 0.27020510536578747599d0, &
         &-0.00000000003097140650d0, 0.00000000010275535029d0,&
         & 0.00000000507546062007d0, -0.00000001697584561546d0, &
         &-0.00000058222618994516d0, 0.00000192483899143098d0,&
         & 0.00004338881805420200d0, -0.00013714074807064326d0, &
         &-0.00184722933402281351d0, 0.00517360719982082990d0,&
         & 0.03611657815299529568d0, -0.07491163418624572802d0, &
         &-0.20317989938720766824d0, 0.17121062620272384486d0, &
         &-0.00000000002289344046d0, -0.00000000027762757821d0,&
         & 0.00000000391330214230d0, 0.00000003673304032817d0, &
         &-0.00000047258696909223d0, -0.00000323109113206792d0,&
         & 0.00003749284946160478d0, 0.00017039067180362504d0, &
         &-0.00172637004639815403d0, -0.00454152531086626043d0,&
         & 0.03717220530377418124d0, 0.04489395902785574534d0, &
         &-0.23370422835726857839d0, -0.06753037249787639679d0,&
         & 0.00000000000473299038d0, -0.00000000040489249115d0, &
         &-0.00000000056747261733d0, 0.00000005656135976496d0,&
         & 0.00000003564150059329d0, -0.00000534821220288168d0,&
         & 0.00000007581385461348d0, 0.00031190256463318867d0, &
         &-0.00014634564376714538d0, -0.00958213551550479850d0,&
         & 0.00624681472076521329d0, 0.11513529702572439703d0, &
         &-0.05794254714300082167d0, -0.22523211169118786537d0,&
         & 0.00000000002732444065d0, -0.00000000017726475495d0, &
         &-0.00000000437562701999d0, 0.00000002681536885560d0,&
         & 0.00000048799100503315d0, -0.00000276490187540644d0, &
         &-0.00003513357925824063d0, 0.00017693155138571443d0,&
         & 0.00144533233498513085d0, -0.00599106092630544975d0, &
         &-0.02759437856689911694d0, 0.07945362316083459396d0,&
         & 0.15383825653750118008d0, -0.17121430684466928734d0 /)
    real(fdp), dimension(0:51), parameter :: d = (/&
         & 1.059601355592185731d-14, -2.71150591218550377d-13,&
         & 8.6514809056201638d-12, -4.6264028554286627d-10,&
         & 5.0815403835647104d-8, -1.76722552048141208d-5,&
         & 0.16286750396763997378d0, 2.949651820598278873d-13, &
         &-8.818215611676125741d-12, 3.571119876162253451d-10, &
         &-2.631924120993717060d-8, 4.709502795656698909d-6, &
         &-5.208333333333283282d-3, 7.18344107717531977d-15, &
         &-2.51623725588410308d-13, 8.6017784918920604d-12, &
         &-4.6256876614290359d-10, 5.0815343220437937d-8, &
         &-1.76722551764941970d-5, 0.16286750396763433767d0,&
         & 2.2327570859680094777d-13, -8.464594853517051292d-12,&
         & 3.563766464349055183d-10, -2.631843986737892965d-8,&
         & 4.709502342288659410d-6, -5.2083333332278466225d-3,&
         & 5.15413392842889366d-15, -2.27740238380640162d-13,&
         & 8.4827767197609014d-12, -4.6224753682737618d-10,&
         & 5.0814848128929134d-8, -1.76722547638767480d-5,&
         & 0.16286750396748926663d0, 1.7316195320192170887d-13, &
         &-7.971122772293919646d-12, 3.544039469911895749d-10, &
         &-2.631443902081701081d-8, 4.709498228695400603d-6, &
         &-5.2083333315143653610d-3, 3.84653681453798517d-15, &
         &-2.04464520778789011d-13, 8.3089298605177838d-12, &
         &-4.6155016158412096d-10, 5.0813263696466650d-8, &
         &-1.76722528311426167d-5, 0.16286750396650065930d0,&
         & 1.3797879972460878797d-13, -7.448089381011684812d-12,&
         & 3.512733797106959780d-10, -2.630500895563592722d-8,&
         & 4.709483934775839193d-6, -5.2083333227940760113d-3 /)

    if (x  <  0.85d0) then
       t = x * x
       y = ((((((a(0) * t + a(1)) * t + a(2)) * t + a(3)) * t +&
            & a(4)) * t + a(5)) * t + a(6)) * t + a(7)
       y = ((((((a(8) * t + a(9)) * t + a(10)) * t + a(11)) * t +&
            & a(12)) * t + a(13)) * t + a(14)) * t + a(15) + y *&
            & log(x)
    else if (x  <  4.5d0) then
       t = x - 4 / x
       k = int(t + 4)
       t = t - (k - 3.5d0)
       k = k * 14
       y = ((((((((((((b(k) * t + b(k + 1)) * t + b(k + 2)) * t +&
            & b(k + 3)) * t + b(k + 4)) * t + b(k + 5)) * t + b(k&
            & + 6)) * t + b(k + 7)) * t + b(k + 8)) * t + b(k +&
            & 9)) * t + b(k + 10)) * t + b(k + 11)) * t + b(k +&
            & 12)) * t + b(k + 13)
    else if (x  <  12.5d0) then
       k = int(x)
       t = x - (k + 0.5d0)
       k = 14 * (k - 4)
       y = ((((((((((((c(k) * t + c(k + 1)) * t + c(k + 2)) * t +&
            & c(k + 3)) * t + c(k + 4)) * t + c(k + 5)) * t + c(k&
            & + 6)) * t + c(k + 7)) * t + c(k + 8)) * t + c(k +&
            & 9)) * t + c(k + 10)) * t + c(k + 11)) * t + c(k +&
            & 12)) * t + c(k + 13)
    else
       v = 24 / x
       t = v * v
       k = 13 * (int(t))
       y = ((((((d(k) * t + d(k + 1)) * t + d(k + 2)) * t + d(k +&
            & 3)) * t + d(k + 4)) * t + d(k + 5)) * t + d(k + 6))&
            & * sqrt(v)
       theta = (((((d(k + 7) * t + d(k + 8)) * t + d(k + 9)) * t &
            &+ d(k + 10)) * t + d(k + 11)) * t + d(k + 12)) * v -&
            & pi4
       y = y * sin(x + theta)
    end if

    f = y
  end function f_besy0
  


  recursive function f_besy1(x) result(f)
    !
    ! Bessel function 2nd kind, order 1
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f
    

    real(fdp) :: t, v, y, theta
    integer :: k
    real(fdp), dimension(0:15), parameter :: a = (/ &
         &-0.00000000000009464990d0, 0.00000000002141432796d0, &
         &-0.00000000359779443470d0, 0.00000043173541397185d0, &
         &-0.00003453883313621946d0, 0.00165786399054057257d0, &
         &-0.03978873577297383381d0, 0.31830988618379067154d0, &
         &-0.00000000005571931463d0, 0.00000000893098810365d0, &
         &-0.00000099267351748541d0, 0.00007164268731546281d0, &
         &-0.00295530533604595764d0, 0.05434868816050718549d0, &
         &-0.19605709064623884320d0, -0.63661977236758134367d0 /)
    real(fdp), dimension(0:111), parameter :: b = (/&
         & 0.00000000002553685555d0, 0.00000000007384080449d0, &
         &-0.00000000029605446927d0, -0.00000000520817391060d0, &
         &-0.00000003676554001376d0, -0.00000015326525336270d0, &
         &-0.00000013967947147386d0, 0.00000388730181348638d0,&
         & 0.00004052240780599446d0, 0.00025133005806772436d0,&
         & 0.00109380629039004777d0, 0.00298157276348996763d0,&
         & 0.00047188983238582062d0, -0.19643830264444486066d0,&
         & 0.00000000001056916126d0, 0.00000000062160122556d0,&
         & 0.00000000385741197273d0, 0.00000000749146940600d0, &
         &-0.00000006306874425799d0, -0.00000068698336577990d0, &
         &-0.00000329785491054157d0, -0.00000557893716372629d0,&
         & 0.00004634569030169549d0, 0.00049086790937752985d0,&
         & 0.00256488729380268318d0, 0.00822571374211646414d0,&
         & 0.01094512145046992425d0, -0.19159562910398869317d0, &
         &-0.00000000040211480479d0, -0.00000000229721816531d0,&
         & 0.00000000062931855684d0, 0.00000006509879184236d0,&
         & 0.00000031589903600713d0, -0.00000003093448289001d0, &
         &-0.00000845532004628483d0, -0.00004930719004112743d0, &
         &-0.00009865914153735679d0, 0.00047059803263198826d0,&
         & 0.00472271367216891688d0, 0.01915518502418162917d0,&
         & 0.03722391361384158740d0, -0.16933230788966878110d0,&
         & 0.00000000110728466147d0, 0.00000000205608441891d0, &
         &-0.00000002536199750468d0, -0.00000011751665325851d0,&
         & 0.00000034045343850770d0, 0.00000436083390507752d0,&
         & 0.00000834725302600388d0, -0.00007154971448642437d0, &
         &-0.00051956165332380432d0, -0.00100811886858456098d0,&
         & 0.00436854257164588225d0, 0.03425644258611852849d0,&
         & 0.09073971172014934990d0, -0.10791597118914788657d0, &
         &-0.00000000117749768422d0, 0.00000000168403859797d0,&
         & 0.00000003471030171126d0, -0.00000006227816248757d0, &
         &-0.00000118405395385178d0, 0.00000014478514453927d0,&
         & 0.00003630744090017806d0, 0.00010481065165516454d0, &
         &-0.00052483993430622472d0, -0.00407014291708523893d0, &
         &-0.00574311514714191566d0, 0.03534768318503458091d0,&
         & 0.16539360869993020163d0, 0.01986240425870529979d0,&
         & 0.00000000040080505520d0, -0.00000000336044652853d0, &
         &-0.00000000092387130705d0, 0.00000017840823064933d0, &
         &-0.00000022643988836534d0, -0.00000808896250262948d0, &
         &-0.00000169403516892975d0, 0.00026617475073041589d0,&
         & 0.00072562705080786576d0, -0.00399208199997748377d0, &
         &-0.02399816920965700841d0, -0.00925574744557542440d0,&
         & 0.20082814022820666990d0, 0.21040564991987672576d0,&
         & 0.00000000006692975871d0, 0.00000000016722027050d0, &
         &-0.00000001433097981251d0, 0.00000005455077276306d0,&
         & 0.00000104954046999028d0, -0.00000339572219821681d0, &
         &-0.00005541577161092338d0, 0.00004344113487321887d0,&
         & 0.00184798594297387512d0, 0.00301041508468150450d0, &
         &-0.02789739141907399514d0, -0.09421628519129373841d0,&
         & 0.09950212445529771576d0, 0.37496975838746404080d0,&
         & 0.00000000004528583204d0, 0.00000000066861092459d0, &
         &-0.00000000924648168570d0, -0.00000008002058622489d0,&
         & 0.00000087578676979688d0, 0.00000628456150216647d0, &
         &-0.00004276239654834506d0, -0.00034630842915911896d0,&
         & 0.00089431202169021474d0, 0.01086380107342193819d0,&
         & 0.00145563493724746238d0, -0.14193383817937981478d0, &
         &-0.15148562296986494725d0, 0.35720232689066512346d0 /)
    real(fdp), dimension(0:125), parameter :: c= (/&
         & 0.00000000032016204980d0, -0.00000000177431944127d0,&
         & 0.00000001367520951871d0, -0.00000006171755212447d0, &
         &-0.00000043602779932091d0, 0.00000179827173442894d0,&
         & 0.00004990624965075363d0, -0.00018618453816525107d0, &
         &-0.00206578613620093339d0, 0.00640292258731756063d0,&
         & 0.04663627606907821596d0, -0.11400070635033154332d0, &
         &-0.26159330264498333979d0, 0.30099732306965462346d0, &
         &-0.00000000000059774708d0, -0.00000000038827648808d0,&
         & 0.00000000527269271362d0, 0.00000002971059814012d0, &
         &-0.00000052701406514027d0, -0.00000320369808906187d0,&
         & 0.00004475144442564492d0, 0.00016907703884518866d0, &
         &-0.00209937055264007564d0, -0.00491023437935628693d0,&
         & 0.04967175540701545140d0, 0.04195559522923170315d0, &
         &-0.33516091307165838036d0, -0.02375823895638961835d0,&
         & 0.00000000000283380103d0, -0.00000000044715450235d0,&
         & 0.00000000017695199610d0, 0.00000006053379840070d0, &
         &-0.00000002888550814408d0, -0.00000593780185020349d0,&
         & 0.00000514027455732102d0, 0.00035667482807472357d0, &
         &-0.00038132729962607730d0, -0.01158750857180657934d0,&
         & 0.01376562658346365929d0, 0.14388461069626297876d0, &
         &-0.13107454661755534290d0, -0.27409127395927545296d0,&
         & 0.00000000002683697058d0, -0.00000000024302161327d0, &
         &-0.00000000430151236236d0, 0.00000003557058903199d0,&
         & 0.00000049328952592926d0, -0.00000365732389065382d0, &
         &-0.00003642356546467712d0, 0.00023636175253326716d0,&
         & 0.00154549935517002108d0, -0.00837084637397210913d0, &
         &-0.02941202173717025197d0, 0.11713664042453548043d0,&
         & 0.15186375421302413107d0, -0.25912851048611625179d0,&
         & 0.00000000002914434604d0, 0.00000000015185121893d0, &
         &-0.00000000488084673389d0, -0.00000001936460341189d0,&
         & 0.00000057968735781303d0, 0.00000158978381365925d0, &
         &-0.00004521858314881102d0, -0.00007426575562367423d0,&
         & 0.00206293925392279862d0, 0.00143934355204571641d0, &
         &-0.04414786410004317489d0, -0.00317227451807568106d0,&
         & 0.27328377353032129600d0, -0.02616867939853747003d0,&
         & 0.00000000000644550655d0, 0.00000000040249587279d0, &
         &-0.00000000123789833341d0, -0.00000005582997578770d0,&
         & 0.00000016975984093144d0, 0.00000524003568623094d0, &
         &-0.00001539871212028104d0, -0.00030372172637669852d0,&
         & 0.00082284448843625174d0, 0.00923614667011392611d0, &
         &-0.02069442879928366384d0, -0.10834973445898588438d0,&
         & 0.14982326837249145873d0, 0.20317989938720766823d0, &
         &-0.00000000002115691215d0, 0.00000000029752035314d0,&
         & 0.00000000334739862269d0, -0.00000004304625868398d0, &
         &-0.00000036733494871207d0, 0.00000425328270524107d0,&
         & 0.00002584872967637478d0, -0.00026244994622929809d0, &
         &-0.00102234403086242668d0, 0.00863185023199066935d0,&
         & 0.01816610124346617164d0, -0.11151661591132254182d0, &
         &-0.08978791805571149950d0, 0.23370422835726857838d0, &
         &-0.00000000002957977086d0, -0.00000000006150420279d0,&
         & 0.00000000488089472193d0, 0.00000000624218182859d0, &
         &-0.00000056561995267846d0, -0.00000032077350100271d0,&
         & 0.00004278569848964830d0, -0.00000053069698280036d0, &
         &-0.00187141538785600230d0, 0.00073172821883575326d0,&
         & 0.03832854206202077374d0, -0.01874044416229564036d0, &
         &-0.23027059405144880640d0, 0.05794254714300082167d0, &
         &-0.00000000001203179184d0, -0.00000000035509887475d0,&
         & 0.00000000213620090322d0, 0.00000004813181550803d0, &
         &-0.00000026815627351126d0, -0.00000439191902440609d0,&
         & 0.00002211921535574542d0, 0.00024593505480524700d0, &
         &-0.00106158930833741898d0, -0.00722666167492552731d0,&
         & 0.02396424370522244155d0, 0.08278313570069734843d0, &
         &-0.15890724632166919294d0, -0.15383825653750118007d0 /)
    real(fdp), dimension(0:51), parameter :: d = (/ &
         &-1.272346002224188092d-14, 3.370464692346669075d-13, &
         &-1.144940314335484869d-11, 6.863141561083429745d-10, &
         &-9.491933932960924159d-8, 5.301676561445687562d-5,&
         & 0.1628675039676399740d0, -3.652982212914147794d-13,&
         & 1.151126750560028914d-11, -5.165585095674343486d-10,&
         & 4.657991250060549892d-8, -1.186794704692706504d-5,&
         & 1.562499999999994026d-2, -8.713069680903981555d-15,&
         & 3.140780373478474935d-13, -1.139089186076256597d-11,&
         & 6.862299023338785566d-10, -9.491926788274594674d-8,&
         & 5.301676558106268323d-5, 0.1628675039676466220d0, &
         &-2.792555727162752006d-13, 1.108650207651756807d-11, &
         &-5.156745588549830981d-10, 4.657894859077370979d-8, &
         &-1.186794650130550256d-5, 1.562499999987299901d-2, &
         &-6.304859171204770696d-15, 2.857249044208791652d-13, &
         &-1.124956921556753188d-11, 6.858482894906716661d-10, &
         &-9.491867953516898460d-8, 5.301676509057781574d-5,&
         & 0.1628675039678191167d0, -2.185193490132496053d-13,&
         & 1.048820673697426074d-11, -5.132819367467680132d-10,&
         & 4.657409437372994220d-8, -1.186794150862988921d-5,&
         & 1.562499999779270706d-2, -4.740417209792009850d-15,&
         & 2.578715253644144182d-13, -1.104148898414138857d-11,&
         & 6.850134201626289183d-10, -9.491678234174919640d-8,&
         & 5.301676277588728159d-5, 0.1628675039690033136d0, &
         &-1.755122057493842290d-13, 9.848723331445182397d-12, &
         &-5.094535425482245697d-10, 4.656255982268609304d-8, &
         &-1.186792402114394891d-5, 1.562499998712198636d-2 /)

    if (x  <  0.85d0) then
       t = x * x
       y = (((((((a(0) * t + a(1)) * t + a(2)) * t + a(3)) * t +&
            & a(4)) * t + a(5)) * t + a(6)) * t + a(7)) * x
       y = (((((((a(8) * t + a(9)) * t + a(10)) * t + a(11)) * t +&
            & a(12)) * t + a(13)) * t + a(14)) * t + a(15)) / x + y *&
            & log(x)
    else if (x  <  4.15d0) then
       v = 4 / x
       t = x - v
       k = int(t + 4)
       t = t - (k - 3.5d0)
       k = k * 14
       y = (((((((((((((b(k) * t + b(k + 1)) * t + b(k + 2)) * t +&
            & b(k + 3)) * t + b(k + 4)) * t + b(k + 5)) * t + b(k +&
            & 6)) * t + b(k + 7)) * t + b(k + 8)) * t + b(k + 9)) * t&
            & + b(k + 10)) * t + b(k + 11)) * t + b(k + 12)) * t +&
            & b(k + 13)) * v
    else if (x  <  12.5d0) then
       k = int(x)
       t = x - (k + 0.5d0)
       k = 14 * (k - 4)
       y = ((((((((((((c(k) * t + c(k + 1)) * t + c(k + 2)) * t + c(k&
            & + 3)) * t + c(k + 4)) * t + c(k + 5)) * t + c(k + 6)) *&
            & t + c(k + 7)) * t + c(k + 8)) * t + c(k + 9)) * t + c(k&
            & + 10)) * t + c(k + 11)) * t + c(k + 12)) * t + c(k + 13)
    else
       v = 24 / x
       t = v * v
       k = 13 * (int(t))
       y = ((((((d(k) * t + d(k + 1)) * t + d(k + 2)) * t + d(k + 3))&
            & * t + d(k + 4)) * t + d(k + 5)) * t + d(k + 6)) *&
            & sqrt(v)
       theta = (((((d(k + 7) * t + d(k + 8)) * t + d(k + 9)) * t +&
            & d(k + 10)) * t + d(k + 11)) * t + d(k + 12)) * v - pi4
       y = y * (-cos(x + theta))
    end if

    f = y

  end function f_besy1



  recursive function f_besyn(n, x) result(f)
    !
    ! Bessel function 2nd kind, order n
    !

    real(fdp), intent(in) :: n, x
    real(fdp) :: f
    
    integer :: j, nn
    real(fdp) :: by, bym, byp, tox

    nn = nint(n)
    if (nn == 0) then
       f = f_besy0(x)
    else if (nn == 1) then
       f = f_besy1(x)
    else
       tox = 2.0d0/x
       by = f_besy1(x)
       bym = f_besy0(x)
       do j = 1, nn-1
          byp = j*tox*by-bym
          bym = by
          by = byp
       end do
       f = by
    end if
  end function f_besyn

  
  
  recursive function f_besi0(x) result(f)
    !
    ! Modified bessel function 1st kind, order 0
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp), dimension(0:64), parameter :: a= (/&
         & 8.5246820682016865877d-11, 2.5966600546497407288d-9,&
         & 7.9689994568640180274d-8, 1.9906710409667748239d-6,&
         & 4.0312469446528002532d-5, 6.4499871606224265421d-4,&
         & 7.9012345761930579108d-3, 7.1111111109207045212d-2,&
         & 4.4444444444472490900d-1, 1.7777777777777532045d0,&
         & 4.0000000000000011182d0, 3.9999999999999999800d0,&
         & 1.0000000000000000001d0, 1.1520919130377195927d-10,&
         & 2.2287613013610985225d-9, 8.1903951930694585113d-8,&
         & 1.9821560631611544984d-6, 4.0335461940910133184d-5,&
         & 6.4495330974432203401d-4, 7.9013012611467520626d-3,&
         & 7.1111038160875566622d-2, 4.4444450319062699316d-1,&
         & 1.7777777439146450067d0, 4.0000000132337935071d0,&
         & 3.9999999968569015366d0, 1.0000000003426703174d0,&
         & 1.5476870780515238488d-10, 1.2685004214732975355d-9,&
         & 9.2776861851114223267d-8, 1.9063070109379044378d-6,&
         & 4.0698004389917945832d-5, 6.4370447244298070713d-4,&
         & 7.9044749458444976958d-3, 7.1105052411749363882d-2,&
         & 4.4445280640924755082d-1, 1.7777694934432109713d0,&
         & 4.0000055808824003386d0, 3.9999977081165740932d0,&
         & 1.0000004333949319118d0, 2.0675200625006793075d-10, &
         &-6.1689554705125681442d-10, 1.2436765915401571654d-7,&
         & 1.5830429403520613423d-6, 4.2947227560776583326d-5,&
         & 6.3249861665073441312d-4, 7.9454472840953930811d-3,&
         & 7.0994327785661860575d-2, 4.4467219586283000332d-1,&
         & 1.7774588182255374745d0, 4.0003038986252717972d0,&
         & 3.9998233869142057195d0, 1.0000472932961288324d0,&
         & 2.7475684794982708655d-10, -3.8991472076521332023d-9,&
         & 1.9730170483976049388d-7, 5.9651531561967674521d-7,&
         & 5.1992971474748995357d-5, 5.7327338675433770752d-4,&
         & 8.2293143836530412024d-3, 6.9990934858728039037d-2,&
         & 4.4726764292723985087d-1, 1.7726685170014087784d0,&
         & 4.0062907863712704432d0, 3.9952750700487845355d0,&
         & 1.0016354346654179322d0 /)
    real(fdp), dimension(0:69), parameter :: b = (/&
         & 6.7852367144945531383d-8, 4.6266061382821826854d-7,&
         & 6.9703135812354071774d-6, 7.6637663462953234134d-5,&
         & 7.9113515222612691636d-4, 7.3401204731103808981d-3,&
         & 6.0677114958668837046d-2, 4.3994941411651569622d-1,&
         & 2.7420017097661750609d0, 14.289661921740860534d0,&
         & 59.820609640320710779d0, 188.78998681199150629d0,&
         & 399.87313678256011180d0, 427.56411572180478514d0,&
         & 1.8042097874891098754d-7, 1.2277164312044637357d-6,&
         & 1.8484393221474274861d-5, 2.0293995900091309208d-4,&
         & 2.0918539850246207459d-3, 1.9375315654033949297d-2,&
         & 1.5985869016767185908d-1, 1.1565260527420641724d0,&
         & 7.1896341224206072113d0, 37.354773811947484532d0,&
         & 155.80993164266268457d0, 489.52113711585409180d0,&
         & 1030.9147225169564806d0, 1093.5883545113746958d0,&
         & 4.8017305613187493564d-7, 3.2613178439123800740d-6,&
         & 4.9073137508166159639d-5, 5.3806506676487583755d-4,&
         & 5.5387918291051866561d-3, 5.1223717488786549025d-2,&
         & 4.2190298621367914765d-1, 3.0463625987357355872d0,&
         & 18.895299447327733204d0, 97.915189029455461554d0,&
         & 407.13940115493494659d0, 1274.3088990480582632d0,&
         & 2670.9883037012547506d0, 2815.7166284662544712d0,&
         & 1.2789926338424623394d-6, 8.6718263067604918916d-6,&
         & 1.3041508821299929489d-4, 1.4282247373727478920d-3,&
         & 1.4684070635768789378d-2, 1.3561403190404185755d-1,&
         & 1.1152592585977393953d0, 8.0387088559465389038d0,&
         & 49.761318895895479206d0, 257.26842323135291380d0,&
         & 1066.8543146269566231d0, 3328.3874581009636362d0,&
         & 6948.8586598121634874d0, 7288.4893398212481055d0,&
         & 3.4093503681970328930d-6, 2.3079025203103376076d-5,&
         & 3.4691373283901830239d-4, 3.7949949772229085450d-3,&
         & 3.8974209677945602145d-2, 3.5949483804148783710d-1,&
         & 2.9522878893539528226d0, 21.246564609514287056d0,&
         & 131.28727387146173141d0, 677.38107093296675421d0,&
         & 2802.3724744545046518d0, 8718.5731420798254081d0,&
         & 18141.348781638832286d0, 18948.925349296308859d0 /)
    real(fdp), dimension(0:44), parameter :: c = (/&
         & 2.5568678676452702768d-15, 3.0393953792305924324d-14,&
         & 6.3343751991094840009d-13, 1.5041298011833009649d-11,&
         & 4.4569436918556541414d-10, 1.7463930514271679510d-8,&
         & 1.0059224011079852317d-6, 1.0729838945088577089d-4,&
         & 5.1503226936425277380d-2, 5.2527963991711562216d-15,&
         & 7.2021184814210056410d-15, 7.2561421229904797156d-13,&
         & 1.4823121466731042510d-11, 4.4602670450376245434d-10,&
         & 1.7463600061788679671d-8, 1.0059226091322347560d-6,&
         & 1.0729838937545111487d-4, 5.1503226936437300716d-2,&
         & 1.3365917359358069908d-14, -1.2932643065888544835d-13,&
         & 1.7450199447905602915d-12, 1.0419051209056979788d-11,&
         & 4.5804788198059832600d-10, 1.7442405450073548966d-8,&
         & 1.0059461453281292278d-6, 1.0729837434500161228d-4,&
         & 5.1503226940658446941d-2, 5.3771611477352308649d-14, &
         &-1.1396193006413731702d-12, 1.2858641335221653409d-11, &
         &-5.9802086004570057703d-11, 7.3666894305929510222d-10,&
         & 1.6731837150730356448d-8, 1.0070831435812128922d-6,&
         & 1.0729733111203704813d-4, 5.1503227360726294675d-2,&
         & 3.7819492084858931093d-14, -4.8600496888588034879d-13,&
         & 1.6898350504817224909d-12, 4.5884624327524255865d-11,&
         & 1.2521615963377513729d-10, 1.8959658437754727957d-8,&
         & 1.0020716710561353622d-6, 1.0730371198569275590d-4,&
         & 5.1503223833002307750d-2 /)
    real(fdp) :: w, t, y
    integer :: k

    w = abs(x)

    if (w  <  8.5d0) then
       t = w * w * 0.0625d0
       k = 13 * (int(t))
       y = (((((((((((a(k) * t + a(k + 1)) * t + a(k + 2)) * t + a(k &
            &+ 3)) * t + a(k + 4)) * t + a(k + 5)) * t + a(k + 6)) *&
            & t + a(k + 7)) * t + a(k + 8)) * t + a(k + 9)) * t + a(k&
            & + 10)) * t + a(k + 11)) * t + a(k + 12)
    else if (w  <  12.5d0) then
       k = int(w)
       t = w - k
       k = 14 * (k - 8)
       y = ((((((((((((b(k) * t + b(k + 1)) * t + b(k + 2)) * t + b(k&
            & + 3)) * t + b(k + 4)) * t + b(k + 5)) * t + b(k + 6)) *&
            & t + b(k + 7)) * t + b(k + 8)) * t + b(k + 9)) * t + b(k&
            & + 10)) * t + b(k + 11)) * t + b(k + 12)) * t + b(k + 13)
    else
       t = 60 / w
       k = 9 * (int(t))
       y = ((((((((c(k) * t + c(k + 1)) * t + c(k + 2)) * t + c(k +&
            & 3)) * t + c(k + 4)) * t + c(k + 5)) * t + c(k + 6)) * t&
            & + c(k + 7)) * t + c(k + 8)) * sqrt(t) * exp(w)
    end if

    f = y
  end function f_besi0



  recursive function f_besi1(x) result(f)
    !
    ! Modified bessel function 1st kind, order 1
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp), dimension(0:59), parameter :: a = (/&
         & 1.2787464404046789181d-10, 3.5705860060088241077d-9,&
         & 9.9611537619347335040d-8, 2.2395070088633043177d-6,&
         & 4.0312466928887462346d-5, 5.6437387840203722356d-4,&
         & 5.9259259312934746096d-3, 4.4444444443499008870d-2,&
         & 2.2222222222232042719d-1, 6.6666666666666139867d-1,&
         & 1.0000000000000001106d0, 4.9999999999999999962d-1,&
         & 1.7281952384448634449d-10, 3.0647204559976390130d-9,&
         & 1.0237662138842827028d-7, 2.2299494417341498163d-6,&
         & 4.0335364374929326943d-5, 5.6433440269141349899d-4,&
         & 5.9259754885893798654d-3, 4.4444399410880397870d-2,&
         & 2.2222225112835026730d-1, 6.6666665422146063244d-1,&
         & 1.0000000032274936821d0, 4.9999999961866867205d-1,&
         & 2.3216048939948030996d-10, 1.7443372702334489579d-9,&
         & 1.1596478963485415499d-7, 2.1446755518623035147d-6,&
         & 4.0697440347437076195d-5, 5.6324394900433192204d-4,&
         & 5.9283484996093060678d-3, 4.4440673899150997921d-2,&
         & 2.2222638016852657860d-1, 6.6666358151576732094d-1,&
         & 1.0000013834029985337d0, 4.9999971643129650249d-1,&
         & 3.1013758938255172562d-10, -8.4813676145611694984d-10,&
         & 1.5544980187411802596d-7, 1.7811109378708045726d-6,&
         & 4.2945322199060856985d-5, 5.5344850176852353639d-4,&
         & 5.9590327716950614802d-3, 4.4371611097707060659d-2,&
         & 2.2233578241986401111d-1, 6.6654747300463315310d-1,&
         & 1.0000756505206705927d0, 4.9997803664415994554d-1,&
         & 4.1214758313965020365d-10, -5.3613317735347429440d-9,&
         & 2.4661360807517345161d-7, 6.7144593918926723203d-7,&
         & 5.1988027944945587571d-5, 5.0165568586065803067d-4,&
         & 6.1717530047005289953d-3, 4.3745229577317251404d-2,&
         & 2.2363147971477747996d-1, 6.6475469131117660240d-1,&
         & 1.0015686689447547657d0, 4.9941120439785391891d-1 /)
    real(fdp), dimension(0:69), parameter :: b = (/&
         & 6.6324787943143095845d-8, 4.5125928898466638619d-7,&
         & 6.7937793134877246623d-6, 7.4580507871505926302d-5,&
         & 7.6866382927334005919d-4, 7.1185174803491859307d-3,&
         & 5.8721838073486424416d-2, 4.2473949281714196041d-1,&
         & 2.6396965606282079123d0, 13.710008536637016903d0,&
         & 57.158647688180932003d0, 179.46182892089389037d0,&
         & 377.57997362398478619d0, 399.87313678256009819d0,&
         & 1.7652713206027939711d-7, 1.1988179244834708057d-6,&
         & 1.8037851545747139231d-5, 1.9775785516370314656d-4,&
         & 2.0354870702829387283d-3, 1.8822164191032253600d-2,&
         & 1.5500485219010424263d-1, 1.1190100010560573210d0,&
         & 6.9391565185406617552d0, 35.948170579648649345d0,&
         & 149.41909525103032616d0, 467.42979492780642582d0,&
         & 979.04227423171290408d0, 1030.9147225169564443d0,&
         & 4.7022299276154507603d-7, 3.1878571710170115972d-6,&
         & 4.7940153875711448496d-5, 5.2496623508411440227d-4,&
         & 5.3968661134780824779d-3, 4.9837081920693776234d-2,&
         & 4.0979593830387765545d-1, 2.9533186922862948404d0,&
         & 18.278176130722516369d0, 94.476497150189121070d0,&
         & 391.66075612645333624d0, 1221.4182034643210345d0,&
         & 2548.6177980961291004d0, 2670.9883037012546541d0,&
         & 1.2535083724002034147d-6, 8.4845871420655708250d-6,&
         & 1.2753227372734042108d-4, 1.3950105363562648921d-3,&
         & 1.4325473993765291906d-2, 1.3212452778932829125d-1,&
         & 1.0849287786885151432d0, 7.8068089156260172673d0,&
         & 48.232254570679165833d0, 248.80659424902394371d0,&
         & 1029.0736929484210803d0, 3200.5629438795801652d0,&
         & 6656.7749162019607914d0, 6948.8586598121632302d0,&
         & 3.3439394490599745013d-6, 2.2600596902211837757d-5,&
         & 3.3955927589987356838d-4, 3.7105306061050972474d-3,&
         & 3.8065263634919156421d-2, 3.5068223415665236079d-1,&
         & 2.8760027832105027316d0, 20.665999500843274339d0,&
         & 127.47939148516390205d0, 656.43636874254000885d0,&
         & 2709.5242837932479920d0, 8407.1174233600734871d0,&
         & 17437.146284159740233d0, 18141.348781638831600d0 /)
    real(fdp), dimension(0:44), parameter :: c = (/ &
         &-2.8849790431465382128d-15, -3.5125350943844774657d-14, &
         &-7.4850867013707419750d-13, -1.8383904048277485153d-11, &
         &-5.7303556446977223342d-10, -2.4449502737311496525d-8, &
         &-1.6765373351766929724d-6, -3.2189516835265773471d-4,&
         & 5.1503226936425277377d-2, -5.8674306822281631119d-15, &
         &-9.4884898451194085565d-15, -8.5033865136600364340d-13, &
         &-1.8142997866945285736d-11, -5.7340238386338193949d-10, &
         &-2.4449138101742183665d-8, -1.6765375646678855842d-6, &
         &-3.2189516826945356325d-4, 5.1503226936412017608d-2, &
         &-1.4723362506764340882d-14, 1.3945147385179042899d-13, &
         &-1.9618041857586930923d-12, -1.3343606394065121821d-11, &
         &-5.8649674606973244159d-10, -2.4426060539669553778d-8, &
         &-1.6765631828366988006d-6, -3.2189515191449587253d-4,&
         & 5.1503226931820146445d-2, -5.8203519372580372987d-14,&
         & 1.2266326995309845825d-12, -1.3921625844526453237d-11,&
         & 6.2228025878281625469d-11, -8.8636681342142794023d-10, &
         &-2.3661241616744818608d-8, -1.6777870960740520557d-6, &
         &-3.2189402882677074318d-4, 5.1503226479551959376d-2, &
         &-4.5801527369223291722d-14, 6.7998819697143727209d-13, &
         &-4.1624857909290468421d-12, -3.2849009406112440998d-11, &
         &-3.2478275690431118270d-10, -2.5739209934053714983d-8, &
         &-1.6730566573215739195d-6, -3.2190010909008684076d-4,&
         & 5.1503229866932077150d-2 /)

    real(fdp) :: w, t, y
    integer :: k

    w = abs(x)

    if (w  <  8.5d0) then
       t = w * w * 0.0625d0
       k = 12 * (int(t))
       y = (((((((((((a(k) * t + a(k + 1)) * t + a(k + 2)) * t + a(k &
            &+ 3)) * t + a(k + 4)) * t + a(k + 5)) * t + a(k + 6)) *&
            & t + a(k + 7)) * t + a(k + 8)) * t + a(k + 9)) * t + a(k&
            & + 10)) * t + a(k + 11)) * w
    else if (w  <  12.5d0) then
       k = int(w)
       t = w - k
       k = 14 * (k - 8)
       y = ((((((((((((b(k) * t + b(k + 1)) * t + b(k + 2)) * t + b(k&
            & + 3)) * t + b(k + 4)) * t + b(k + 5)) * t + b(k + 6)) *&
            & t + b(k + 7)) * t + b(k + 8)) * t + b(k + 9)) * t + b(k&
            & + 10)) * t + b(k + 11)) * t + b(k + 12)) * t + b(k + 13)
    else
       t = 60 / w
       k = 9 * (int(t))
       y = ((((((((c(k) * t + c(k + 1)) * t + c(k + 2)) * t + c(k +&
            & 3)) * t + c(k + 4)) * t + c(k + 5)) * t + c(k + 6)) * t&
            & + c(k + 7)) * t + c(k + 8)) * sqrt(t) * exp(w)
    end if

    if (x  <  0) y = -y

    f = y
  end function f_besi1



  recursive function f_besin(n, x) result(f)
    !
    ! Modified bessel function 1st kind, order n
    !

    real(fdp), intent(in) :: n, x
    real(fdp) :: f

    integer, parameter :: iacc=40, iexp=maxexponent(x)/2
    integer :: j, m, nn
    real(fdp) :: bi, bim, bip, tox

    nn = nint(n)

    if (nn == 0) then
       f = f_besi0(x)
    else if (nn == 1) then
       f = f_besi1(x)
    else
       f=0.0

       if (x*x <= 8.0d0*tiny(x)) return

       tox=2.0d0/abs(x)
       bip=0.0d0
       bi=1.0d0
       m=2*((nn+int(sqrt(real(iacc*nn)))))

       do j=m,1,-1
          bim=bip+j*tox*bi
          bip=bi
          bi=bim

          if (exponent(bi) > iexp) then
             f=scale(f,-iexp)
             bi=scale(bi,-iexp)
             bip=scale(bip,-iexp)
          end if

          if (j == nn) f=bip
       end do

       f=f*f_besi0(x)/bi

       if (x < 0.0 .and. mod(nn,2) == 1) f=-f
    end if
  end function f_besin



  recursive function f_besk0(x) result(f)
    !
    ! Modified bessel function 2nd kind, order 0
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp), dimension(0:15), parameter :: a = (/&
         & 2.4307270476772195953d-12, 4.7091666363785304370d-10,&
         & 6.7816861334344265568d-8, 6.7816840204737508252d-6,&
         & 4.3402777777915334676d-4, 1.5624999999999872796d-2,&
         & 2.5000000000000000448d-1, 9.9999999999999999997d-1,&
         & 6.5878327432224993071d-12, 1.2083308769932888218d-9,&
         & 1.6271062073716412046d-7, 1.4914719278555277887d-5,&
         & 8.4603509071212245667d-4, 2.5248929932162333910d-2,&
         & 2.7898287891460312491d-1, 1.1593151565841244874d-1 /)
    real(fdp), dimension(0:111), parameter :: b = (/ &
         &-4.6430702971053162197d-13, 1.0377936059563728230d-11, &
         &-1.0298475936392057807d-10, 5.3632747492333959219d-10, &
         &-2.1674628861036068105d-10, -2.3316071545820437669d-8,&
         & 2.2557819578691704059d-7, -9.2325694638587080009d-7, &
         &-3.3569097781613661759d-6, 8.7355061305812582974d-5, &
         &-6.8021202111645760475d-4, 2.7434654781323362319d-4,&
         & 1.0031787169953909561d-1, 4.2102443824070833334d-1,&
         & 4.1447451117883103686d-12, -3.4026589638576604315d-11,&
         & 9.3398790624638977468d-12, 1.5184181750799852630d-9, &
         &-1.1364911665083029464d-8, 2.0619457602095915719d-8,&
         & 3.0431018037572243630d-7, -2.9749736264474555510d-6,&
         & 8.0143661611467038568d-6, 8.0937525149549218398d-5, &
         &-1.0356346549612699886d-3, 2.8534806627578638795d-3,&
         & 9.7369634474060441807d-2, 3.2175066577856452683d-1,&
         & 1.1170882570740727520d-13, -8.2865909408297066068d-11,&
         & 9.4656678749191182763d-10, -3.5832019841847883380d-9, &
         &-9.5017955656904252761d-9, 1.5200595674883329093d-7, &
         &-3.8663262571356059980d-7, -3.3350340828235103499d-6,&
         & 2.9359886663960844231d-5, -1.1266401822556801563d-5, &
         &-1.2113572742435576205d-3, 6.3158973673701376253d-3,&
         & 8.8291790250128171341d-2, 2.2833982383240512262d-1, &
         &-3.2880638807053948433d-11, 4.3194884830465283512d-10, &
         &-1.7455089683104033093d-9, -3.2437330799994764516d-9,&
         & 4.7393655539139519778d-8, -1.1929265603456272466d-8, &
         &-1.3177845881013419388d-6, 3.3873375636197969526d-6,&
         & 3.2729835880668256625d-5, -1.8367283883002494561d-4, &
         &-8.2830996454188084408d-4, 9.5512732229514251931d-3,&
         & 7.2233832113719266702d-2, 1.4753187103603405298d-1,&
         & 7.9998492614150860098d-11, -7.0257346702686139490d-10,&
         & 7.8898821627084586270d-10, 1.1294796399671507085d-8, &
         &-1.1360539648638059137d-8, -3.0346309115270564487d-7,&
         & 3.2235585426189451721d-7, 8.3575612102298214948d-6, &
         &-8.5169628089198208211d-6, -2.5740175232173357342d-4,&
         & 1.2462734014689152770d-4, 1.0683232869192203450d-2,&
         & 5.1515690033637395779d-2, 8.5465862953544883657d-2, &
         &-8.6111506537356531608d-11, 5.1862926131024597823d-10,&
         & 7.5884324949371110022d-10, -6.4011975813006767417d-9, &
         &-4.1966181325111763156d-8, 9.1306285446881485314d-8,&
         & 1.3573638315827954034d-6, 4.8683213252735694701d-7, &
         &-3.8805424608710197066d-5, -1.1838986468688980610d-4,&
         & 9.2796213947750964945d-4, 8.9611057737319027776d-3,&
         & 3.1464453915862785606d-2, 4.4267648087536630780d-2,&
         & 4.4400123834164610288d-11, -1.1411233140911074336d-10, &
         &-8.8200670702467059830d-10, -1.9686735373323381456d-9,&
         & 1.9921120728941773855d-8, 1.4543974418584834740d-7,&
         & 1.8238418041265854754d-8, -4.5363700392899066037d-6, &
         &-2.1688068222527688542d-5, 4.5496062166687034700d-5,&
         & 1.0435238076080528284d-3, 5.8374528996419979931d-3,&
         & 1.6611210710425455850d-2, 2.0756008367065750538d-2, &
         &-6.5166519951106397214d-12, -5.8572182858788539580d-11,&
         & 1.5550375065815375404d-10, 1.9526509484993563229d-9,&
         & 9.2637123346818426594d-9, -1.4136471501812055943d-8, &
         &-4.3024895710889717172d-7, -2.3235612243330592076d-6,&
         & 4.0380616133862188804d-7, 9.2783767992909743602d-5,&
         & 7.2964887597817095035d-4, 3.1316245282223273413d-3,&
         & 7.8028233022066428316d-3, 9.0014807263791058095d-3 /)
    real(fdp), dimension(0:134), parameter :: c = (/&
         & 4.5161032649342790231d-11, -4.2774336988557091369d-11,&
         & 6.0998467173896677777d-10, 1.9845167242599996944d-9,&
         & 1.3097678767280215271d-8, 7.4505822268382641286d-8,&
         & 4.2893920879106814989d-7, 2.3900851955655303104d-6,&
         & 1.2533473009382380357d-5, 5.9693359063879871983d-5,&
         & 2.4775070661087304580d-4, 8.5106703131389516508d-4,&
         & 2.2500105115665788755d-3, 4.0446134454521634600d-3,&
         & 3.6910983340425942762d-3, 3.5732826433251464989d-12, &
         &-3.2906649482312266258d-12, 7.0873811190464760555d-11,&
         & 2.9551320580484177120d-10, 2.2776940472505079894d-9,&
         & 1.5175463612815010036d-8, 9.9462487812170164133d-8,&
         & 6.1448757797853901100d-7, 3.4869531882907360750d-6,&
         & 1.7615836644757657443d-5, 7.6373536037879531886d-5,&
         & 2.7098571871205999668d-4, 7.3399047381788927036d-4,&
         & 1.3439197177355085297d-3, 1.2439943280131230863d-3,&
         & 3.6343547836242523646d-13, 9.7997961751276137602d-14,&
         & 1.0184692699811569047d-11, 6.1495184828957652064d-11,&
         & 5.0238328349302602543d-10, 3.7498626376004337661d-9,&
         & 2.6689445483857236307d-8, 1.7591899737346368084d-7,&
         & 1.0486448307010701679d-6, 5.4986458466257148573d-6,&
         & 2.4521456351751345323d-5, 8.8900942259143832228d-5,&
         & 2.4483947714068300190d-4, 4.5418248688489693045d-4,&
         & 4.2479574186923180694d-4, 5.2460389348163395857d-14,&
         & 7.4802063026503503540d-14, 2.0012201610651998417d-12,&
         & 1.4887306044735163359d-11, 1.2946705414232940350d-10,&
         & 1.0391628915892803144d-9, 7.8091180499677328456d-9,&
         & 5.3694223626907660084d-8, 3.3063914804658509029d-7,&
         & 1.7776972424421486506d-6, 8.0833148098458320202d-6,&
         & 2.9755556304448817780d-5, 8.2945928349220642178d-5,&
         & 1.5536921180500112883d-4, 1.4647070522281538711d-4,&
         & 9.7531436733955514559d-15, 2.4084291220447154982d-14,&
         & 4.7654956400897494468d-13, 4.0200949504810597783d-12,&
         & 3.6726577109162191533d-11, 3.0939005665422637601d-10,&
         & 2.4122848979784500179d-9, 1.7071884462645525505d-8,&
         & 1.0752238955654933405d-7, 5.8844190041189462347d-7,&
         & 2.7136083303224014597d-6, 1.0102477728604441135d-5,&
         & 2.8420490721532571809d-5, 5.3637016379451944413d-5,&
         & 5.0881312956459247572d-5, 2.1732049868189377260d-15,&
         & 7.2720052142815590531d-15, 1.2803083795536820100d-13,&
         & 1.1696825543787717167d-12, 1.1083298191597132094d-11,&
         & 9.6536661252658773139d-11, 7.7242553835198536397d-10,&
         & 5.5798366267110575620d-9, 3.5721345296543414370d-8,&
         & 1.9806931547193682466d-7, 9.2312964655319555313d-7,&
         & 3.4666258590861079959d-6, 9.8224698307751177077d-6,&
         & 1.8648773453825584428d-5, 1.7780062316167651812d-5,&
         & 5.5012463763851934112d-16, 2.2254763392767319419d-15,&
         & 3.7187669817701214965d-14, 3.5819585377733489628d-13,&
         & 3.4866061263191556694d-12, 3.1101633450629652910d-11,&
         & 2.5358235662235617663d-10, 1.8597629779492599046d-9,&
         & 1.2052654739462999992d-8, 6.7501417351172136833d-8,&
         & 3.1720052198654584574d-7, 1.1993651363602981832d-6,&
         & 3.4179130317623363474d-6, 6.5208606745808860158d-6,&
         & 6.2430205476536771454d-6, 1.5225407517829491689d-16,&
         & 6.9834820025664405161d-16, 1.1380182837138781431d-14,&
         & 1.1369488761077196511d-13, 1.1291168681618466716d-12,&
         & 1.0250757630526871007d-11, 8.4765287317253141514d-11,&
         & 6.2886627779402596211d-10, 4.1142865598366029316d-9,&
         & 2.3223773435632014408d-8, 1.0985095234166396934d-7,&
         & 4.1766260951820336228d-7, 1.1958609263543792991d-6,&
         & 2.2907574647671878055d-6, 2.2008253973114914005d-6,&
         & 4.4863058691420695911d-17, 2.2437356594371819978d-16,&
         & 3.6107964803015652759d-15, 3.7031193629853392081d-14,&
         & 3.7341552790439784371d-13, 3.4355950129497564468d-12,&
         & 2.8719942600171304499d-11, 2.1499646844509516453d-10,&
         & 1.4171810843455227171d-9, 8.0501118772875784153d-9,&
         & 3.8281889106330295876d-8, 1.4621673458431979989d-7,&
         & 4.2029868696411098586d-7, 8.0785884122023473025d-7,&
         & 7.7845438614204963209d-7 /)
    real(fdp), dimension(0:39), parameter :: d = (/ &
         &-7.9737703860537066166d-14, 1.9543834380466766627d-12, &
         &-4.7230794431646733538d-11, 1.4001773785771252004d-9, &
         &-5.4864553020583098585d-8, 3.1601984250143742772d-6, &
         &-3.3708783204090252161d-4, 1.6180215937964160437d-1, &
         &-5.2593898374798632343d-14, 1.7725913926973236457d-12, &
         &-4.6672234858122387294d-11, 1.3991653503828889207d-9, &
         &-5.4863400156413929639d-8, 3.1601976099900075541d-6, &
         &-3.3708783171335864627d-4, 1.6180215937958433760d-1, &
         &-3.6135496189875398132d-14, 1.5466239429618130284d-12, &
         &-4.5320259146602122624d-11, 1.3945974109459385552d-9, &
         &-5.4853994841172088787d-8, 3.1601858228022739196d-6, &
         &-3.3708782339998302320d-4, 1.6180215937704286491d-1, &
         &-2.5640663123518180635d-14, 1.3288079339404032671d-12, &
         &-4.3368537955908371563d-11, 1.3848103653102203186d-9, &
         &-5.4824335664256344123d-8, 3.1601315173126153586d-6, &
         &-3.3708776779035695640d-4, 1.6180215935248373474d-1, &
         &-1.8678321325292127767d-14, 1.1354310934105733311d-12, &
         &-4.1057197297998608931d-11, 1.3693990961296350970d-9, &
         &-5.4762428935047089835d-8, 3.1599817092775027963d-6, &
         &-3.3708756559715893599d-4, 1.6180215923508144240d-1 /)
    real(fdp) :: t, y
    integer :: k

    if (x  <  0.86d0) then
       t = x * x
       y = ((((((a(0) * t + a(1)) * t + a(2)) * t + a(3)) * t + a(4))&
            & * t + a(5)) * t + a(6)) * t + a(7)
       y = ((((((a(8) * t + a(9)) * t + a(10)) * t + a(11)) * t +&
            & a(12)) * t + a(13)) * t + a(14)) * t + a(15) - y *&
            & log(x)
    else if (x  <  4.15d0) then
       t = x - 5 / x
       k = int(t + 5)
       t = (k - 4) - t
       k = k * 14
       y = ((((((((((((b(k) * t + b(k + 1)) * t + b(k + 2)) * t + b(k&
            & + 3)) * t + b(k + 4)) * t + b(k + 5)) * t + b(k + 6)) *&
            & t + b(k + 7)) * t + b(k + 8)) * t + b(k + 9)) * t + b(k&
            & + 10)) * t + b(k + 11)) * t + b(k + 12)) * t + b(k + 13)
    else if (x  <  12.5d0) then
       k = int(x)
       t = (k + 1) - x
       k = 15 * (k - 4)
       y = (((((((((((((c(k) * t + c(k + 1)) * t + c(k + 2)) * t +&
            & c(k + 3)) * t + c(k + 4)) * t + c(k + 5)) * t + c(k +&
            & 6)) * t + c(k + 7)) * t + c(k + 8)) * t + c(k + 9)) * t&
            & + c(k + 10)) * t + c(k + 11)) * t + c(k + 12)) * t +&
            & c(k + 13)) * t + c(k + 14)
    else
       t = 60 / x
       k = 8 * (int(t))
       y = (((((((d(k) * t + d(k + 1)) * t + d(k + 2)) * t + d(k +&
            & 3)) * t + d(k + 4)) * t + d(k + 5)) * t + d(k + 6)) * t&
            & + d(k + 7)) *     sqrt(t) * exp(-x)
    end if

    f = y
  end function f_besk0



  recursive function f_besk1(x) result(f)
    !
    ! Modified bessel function 2nd kind, order 1
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp), dimension(0:15), parameter :: a = (/&
         & 1.5151605362537935201d-13, 3.3637909513536510350d-11,&
         & 5.6514041131016827202d-9, 6.7816840255069534052d-7,&
         & 5.4253472222259226487d-5, 2.6041666666666637057d-3,&
         & 6.2500000000000000090d-2, 5.0000000000000000000d-1, &
         &-8.9790303384748696588d-11, -1.4029047449249185771d-8, &
         &-1.5592893752540998113d-6, -1.1253607018469017569d-4, &
         &-4.6421827665011579173d-3, -8.5370719728648409609d-2, &
         &-3.0796575782920629660d-1, 1.0000000000000000004d0 /)
    real(fdp), dimension(0:119), parameter :: b = (/ &
         &-9.4055461896630579928d-12, 3.1307934665844664773d-11,&
         & 4.2005295001519243251d-10, -4.1636196779679820012d-9,&
         & 1.4483026181700966164d-8, 1.1661000205428816914d-8, &
         &-3.5023996724943046209d-7, 1.4404279316339005012d-6,&
         & 5.3581564157158242080d-7, -3.5249754038612334639d-5,&
         & 1.7150324075631641453d-4, -4.1276362611239191024d-5, &
         &-4.6943110979636602591d-3, 3.5085369853392357659d-2,&
         & 2.0063574339907819159d-1, 3.3998989888944034586d-11,&
         & 7.1558979072937373055d-11, -2.9226856932927698732d-9,&
         & 1.4591620256525610213d-8, -6.6141635609854161666d-9, &
         &-1.9991101838984472332d-7, 5.9185836628873530572d-7,&
         & 1.9562880347358085687d-6, -1.5814366450418102764d-5,&
         & 7.6791682910944612028d-6, 2.8354678948323983936d-4, &
         &-1.0217932669418690641d-3, -3.2205661865210048433d-3,&
         & 4.3497494842354644077d-2, 1.6110284302315089935d-1, &
         &-2.0933987679665541827d-10, 7.9503322090520447316d-10,&
         & 3.8000150948242575774d-9, -2.3076136195585571309d-8, &
         &-2.3902880302550799653d-8, 3.1914500937804377478d-7,&
         & 3.2639909831082417694d-7, -5.3166994792995439449d-6, &
         &-3.1109524694269240094d-6, 9.2575906966353273247d-5,&
         & 7.5063709094147644746d-7, -1.7416491592625765379d-3,&
         & 1.2138560335171676007d-3, 4.5879687144659643175d-2,&
         & 1.1566544716132846709d-1, 3.1582384905164908749d-10, &
         &-1.9959561818098999516d-9, 8.6959328920030927557d-10,&
         & 1.1642778282445577109d-8, 4.3552264337818440471d-8, &
         &-1.5057982160481803238d-7, -1.0101729117980989857d-6,&
         & 7.7002002510177612013d-7, 1.9580574235590194233d-5,&
         & 1.9358461980242834361d-5, -3.3932339942485532728d-4, &
         &-9.3416673584325090073d-4, 5.5800080455912847227d-3,&
         & 3.8668683062477179235d-2, 7.2651643500517000658d-2, &
         &-1.1554749629758510059d-10, 8.2270678758893273006d-10, &
         &-5.0211156951551538591d-10, -1.4929179050554858361d-9, &
         &-2.7107940791526366702d-8, -4.2204764086705349384d-8,&
         & 3.7253098167927628867d-7, 2.4374697215363361156d-6,&
         & 1.4141942006909768370d-6, -4.8766389019473918231d-5, &
         &-2.1681387247526720978d-4, 2.9325729929653405236d-4,&
         & 6.4087534504827239815d-3, 2.6054628289709454356d-2,&
         & 4.0156431128194184336d-2, 2.5506555170746221691d-11, &
         &-1.3521164018407978152d-10, -8.3281235274106699399d-11, &
         &-9.7764849575562351891d-10, 3.4661828409940354542d-9,&
         & 3.9760633711791357544d-8, 1.5902906645504529930d-7, &
         &-1.4919441249454941275d-7, -5.3779684992094351263d-6, &
         &-2.7513862296246223142d-5, -9.7880089725297162007d-6,&
         & 7.0787668964515789714d-4, 4.6968199862345387583d-3,&
         & 1.4745740181663320127d-2, 2.0048622219583455723d-2, &
         &-3.4824483072529265585d-12, 1.5157161810563380451d-12,&
         & 8.5303859696700686144d-12, 3.3455414203743741076d-10,&
         & 2.0226016353844285376d-9, 5.3128154003266334990d-9, &
         &-3.0799322316418042137d-8, -4.4455408272954712128d-7, &
         &-2.4293274626893384034d-6, -3.2129079340119038354d-6,&
         & 5.9225403683075388850d-5, 5.6822962576781683532d-4,&
         & 2.7152446516406682732d-3, 7.4075873691848838485d-3,&
         & 9.3044450815739269849d-3, -2.7683216166276377232d-13,&
         & 3.1986676777610155465d-12, 9.4142986954031445666d-12,&
         & 6.7934609179456399334d-11, 3.4851529411470029330d-11, &
         &-2.5785248508896551557d-9, -2.8310220027112571258d-8, &
         &-1.6384131113072271115d-7, -3.2521663350596379097d-7,&
         & 4.0381388757622307160d-6, 5.1917606978077281001d-5,&
         & 3.3420027947470126154d-4, 1.3699550623118247094d-3,&
         & 3.4405619148342271096d-3, 4.1042919106665762794d-3 /)
    real(fdp), dimension(0:119), parameter :: c = (/&
         & 4.5281968025889407937d-12, 1.0806749918195271176d-11,&
         & 9.6200972728717669027d-11, 5.7214227063625263650d-10,&
         & 3.6077804282954825099d-9, 2.2465236858536681852d-8,&
         & 1.3676961264308735230d-7, 7.9561767489531997361d-7,&
         & 4.3014380065615550573d-6, 2.0921713905550285590d-5,&
         & 8.8079183950590176926d-5, 3.0549414408830252064d-4,&
         & 8.1295715613927890473d-4, 1.4679809476357079195d-3,&
         & 1.3439197177355090057d-3, 7.6019964430402432637d-13, &
         &-2.2616198599158271190d-13, 1.7904450823779000744d-11,&
         & 9.1467054855312232717d-11, 7.1378582044879519122d-10,&
         & 4.9925255415445769102d-9, 3.3767315471315546644d-8,&
         & 2.1350774539167751457d-7, 1.2314353082655232903d-6,&
         & 6.2918685053670619181d-6, 2.7493229298777000013d-5,&
         & 9.8085825401369821771d-5, 2.6670282677770444935d-4,&
         & 4.8967895428135985381d-4, 4.5418248688489697144d-4,&
         & 9.4180115230375147213d-14, 7.5943117003734061145d-14,&
         & 3.0335730243874287654d-12, 2.0202796115462268051d-11,&
         & 1.6839020189186971198d-10, 1.2907875663127201526d-9,&
         & 9.3547676125865798920d-9, 6.2471974110281880722d-8,&
         & 3.7585985422997380441d-7, 1.9838348288114906484d-6,&
         & 8.8884862203671982034d-6, 3.2333259238682810218d-5,&
         & 8.9266668913380400243d-5, 1.6589185669844051903d-4,&
         & 1.5536921180500113394d-4, 1.5425475332301107271d-14,&
         & 2.8674534590132490434d-14, 6.5078462279160216936d-13,&
         & 5.0939757793961391211d-12, 4.4979837460748975520d-11,&
         & 3.6662925847520171711d-10, 2.7848878755089582413d-9,&
         & 1.9298120059339477820d-8, 1.1950323861976892013d-7,&
         & 6.4513432758147478287d-7, 2.9422095033982461936d-6,&
         & 1.0854433321174584937d-5, 3.0307433185818899481d-5,&
         & 5.6840981443065017850d-5, 5.3637016379451945253d-5,&
         & 3.1077953698439839352d-15, 8.6899496170729520378d-15,&
         & 1.6258562067326054104d-13, 1.4104842571366761537d-12,&
         & 1.3019455544084110747d-11, 1.1070466372863950239d-10,&
         & 8.6890603844230597917d-10, 6.1793722175049967488d-9,&
         & 3.9058865943755615801d-8, 2.1432806981070368523d-7,&
         & 9.9034657762983230155d-7, 3.6925185861895664251d-6,&
         & 1.0399877577259449786d-5, 1.9644939661550210015d-5,&
         & 1.8648773453825584597d-5, 7.2831555285336286457d-16,&
         & 2.6077534095895783532d-15, 4.4881202059263153495d-14,&
         & 4.1674329383944385626d-13, 3.9760100480223728037d-12,&
         & 3.4835976355351183010d-11, 2.7993254212770249700d-10,&
         & 2.0286513276830758107d-9, 1.3018343087118439152d-8,&
         & 7.2315927974997999365d-8, 3.3750708681924201599d-7,&
         & 1.2688020879407355571d-6, 3.5980954090811587848d-6,&
         & 6.8358260635246667316d-6, 6.5208606745808860557d-6,&
         & 1.9026412343503745875d-16, 8.0073765508732553766d-16,&
         & 1.3245754278055523992d-14, 1.2885201653055058502d-13,&
         & 1.2600129301230402587d-12, 1.1283306843147549277d-11,&
         & 9.2261481309646814329d-11, 6.7812033168299846818d-10,&
         & 4.4020645304595102132d-9, 2.4685719238301517679d-8,&
         & 1.1611886719473112509d-7, 4.3940380936523135466d-7,&
         & 1.2529878285546791905d-6, 2.3917218527087570384d-6,&
         & 2.2907574647671878160d-6, 5.3709522135744366512d-17,&
         & 2.5239221050372845433d-16, 4.0933147145899083360d-15,&
         & 4.1152784247617592367d-14, 4.0998840572769381012d-13,&
         & 3.7319354625807158852d-12, 3.0921671702920868014d-11,&
         & 2.2975898538634445343d-10, 1.5049754445782364328d-9,&
         & 8.5030864719789148982d-9, 4.0250559391118423810d-8,&
         & 1.5312755642491878591d-7, 4.3865020375297892208d-7,&
         & 8.4059737392822153101d-7, 8.0785884122023473319d-7 /)
    real(fdp), dimension(0:39), parameter :: d = (/&
         & 9.2371554649979581914d-14, -2.3111336195788410887d-12,&
         & 5.7728710326649832559d-11, -1.8002298130091372598d-9,&
         & 7.6810375010517145638d-8, -5.2669973752193823306d-6,&
         & 1.0112634961227401357d-3, 1.6180215937964160466d-1,&
         & 6.1381146507252683381d-14, -2.1034499679806301862d-12,&
         & 5.7090233460448415278d-11, -1.7990724350642330817d-9,&
         & 7.6809056078388019946d-8, -5.2669964425290062357d-6,&
         & 1.0112634957478283390d-3, 1.6180215937970716383d-1,&
         & 4.2458150578401296419d-14, -1.8435733128339016981d-12,&
         & 5.5534955081564656595d-11, -1.7938162188526358466d-9,&
         & 7.6798230945934117807d-8, -5.2669828728791921259d-6,&
         & 1.0112634861753356559d-3, 1.6180215938263409582d-1,&
         & 3.0314798962267007518d-14, -1.5915009905364214455d-12,&
         & 5.3275907427402047438d-11, -1.7824862013841369751d-9,&
         & 7.6763890356447075810d-8, -5.2669199860465945909d-6,&
         & 1.0112634217687349189d-3, 1.6180215941108227283d-1,&
         & 2.2211515002229271212d-14, -1.3664088221521734796d-12,&
         & 5.0585177270502341602d-11, -1.7645432205894533462d-9,&
         & 7.6691805594577373698d-8, -5.2667455286976269634d-6,&
         & 1.0112631862810974580d-3, 1.6180215954783127877d-1 /)
    real(fdp) :: y, v, t
    integer :: k

    if (x  <  0.8d0) then
       t = x * x
       y = (((((((a(0) * t + a(1)) * t + a(2)) * t + a(3)) * t +&
            & a(4)) * t + a(5)) * t + a(6)) * t + a(7)) * x
       y = (((((((a(8) * t + a(9)) * t + a(10)) * t + a(11)) * t +&
            & a(12)) * t + a(13)) * t + a(14)) * t + a(15)) / x + y *&
            & log(x)
    else if (x  <  5.5d0) then
       v = 3 / x
       t = x - v
       k = int(t + 3)
       t = (k - 2) - t
       k = k * 15
       y = ((((((((((((((b(k) * t + b(k + 1)) * t + b(k + 2)) * t +&
            & b(k + 3)) * t + b(k + 4)) * t + b(k + 5)) * t + b(k +&
            & 6)) * t + b(k + 7)) * t + b(k + 8)) * t + b(k + 9)) * t&
            & + b(k + 10)) * t + b(k + 11)) * t + b(k + 12)) * t +&
            & b(k + 13)) * t + b(k + 14)) * v
    else if (x  <  12.5d0) then
       k = int(x)
       t = (k + 1) - x
       k = 15 * (k - 5)
       y = (((((((((((((c(k) * t + c(k + 1)) * t + c(k + 2)) * t +&
            & c(k + 3)) * t + c(k + 4)) * t + c(k + 5)) * t + c(k +&
            & 6)) * t + c(k + 7)) * t + c(k + 8)) * t + c(k + 9)) * t&
            & + c(k + 10)) * t + c(k + 11)) * t + c(k + 12)) * t +&
            & c(k + 13)) * t + c(k + 14)
    else
       t = 60 / x
       k = 8 * (int(t))
       y = (((((((d(k) * t + d(k + 1)) * t + d(k + 2)) * t + d(k +&
            & 3)) * t + d(k + 4)) * t + d(k + 5)) * t + d(k + 6)) * t&
            & + d(k + 7)) *     sqrt(t) * exp(-x)
    end if

    f = y
  end function f_besk1



  recursive function f_beskn(n, x) result(f)
    !
    ! Modified bessel function 2nd kind, order n
    !

    real(fdp), intent(in) :: n, x
    real(fdp) :: f

    integer :: j, nn
    real(fdp) :: bk,bkm,bkp,tox
    
    nn = nint(n)

    if (nn == 0) then
       f = f_besk0(x)
    else if (nn == 1) then
       f = f_besk1(x)
    else
       tox=2.0d0/x
       bkm=f_besk0(x)
       bk=f_besk1(x)

       do j=1,nn-1
          bkp=bkm+j*tox*bk
          bkm=bk
          bk=bkp
       end do

       f=bk
    end if
  end function f_beskn



  recursive function f_erf(x) result(f)
    !
    ! Error function
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp), dimension(0:64), parameter :: a = (/ 0.00000000005958930743d0, &
         -0.00000000113739022964d0, 0.00000001466005199839d0, -0.00000016350354461960d0, &
         0.00000164610044809620d0, -0.00001492559551950604d0, 0.00012055331122299265d0, &
         -0.00085483269811296660d0, 0.00522397762482322257d0, -0.02686617064507733420d0, &
         0.11283791670954881569d0, -0.37612638903183748117d0, 1.12837916709551257377d0, &
         0.00000000002372510631d0, -0.00000000045493253732d0, 0.00000000590362766598d0, &
         -0.00000006642090827576d0, 0.00000067595634268133d0, -0.00000621188515924000d0, &
         0.00005103883009709690d0, -0.00037015410692956173d0, 0.00233307631218880978d0, &
         -0.01254988477182192210d0, 0.05657061146827041994d0, -0.21379664776456006580d0, &
         0.84270079294971486929d0, 0.00000000000949905026d0, -0.00000000018310229805d0, &
         0.00000000239463074000d0, -0.00000002721444369609d0, 0.00000028045522331686d0, &
         -0.00000261830022482897d0, 0.00002195455056768781d0, -0.00016358986921372656d0, &
         0.00107052153564110318d0, -0.00608284718113590151d0, 0.02986978465246258244d0, &
         -0.13055593046562267625d0, 0.67493323603965504676d0, 0.00000000000382722073d0, &
         -0.00000000007421598602d0, 0.00000000097930574080d0, -0.00000001126008898854d0, &
         0.00000011775134830784d0, -0.00000111992758382650d0, 0.00000962023443095201d0, &
         -0.00007404402135070773d0, 0.00050689993654144881d0, -0.00307553051439272889d0, &
         0.01668977892553165586d0, -0.08548534594781312114d0, 0.56909076642393639985d0, &
         0.00000000000155296588d0, -0.00000000003032205868d0, 0.00000000040424830707d0, &
         -0.00000000471135111493d0, 0.00000005011915876293d0, -0.00000048722516178974d0, &
         0.00000430683284629395d0, -0.00003445026145385764d0, 0.00024879276133931664d0, &
         -0.00162940941748079288d0, 0.00988786373932350462d0, -0.05962426839442303805d0, &
         0.49766113250947636708d0 /)
    real(fdp), dimension(0:64), parameter :: b = (/ -0.00000000029734388465d0, &
         0.00000000269776334046d0, -0.00000000640788827665d0, -0.00000001667820132100d0, &
         -0.00000021854388148686d0, 0.00000266246030457984d0, 0.00001612722157047886d0, &
         -0.00025616361025506629d0, 0.00015380842432375365d0, 0.00815533022524927908d0, &
         -0.01402283663896319337d0, -0.19746892495383021487d0, 0.71511720328842845913d0, &
         -0.00000000001951073787d0, -0.00000000032302692214d0, 0.00000000522461866919d0, &
         0.00000000342940918551d0, -0.00000035772874310272d0, 0.00000019999935792654d0, &
         0.00002687044575042908d0, -0.00011843240273775776d0, -0.00080991728956032271d0, &
         0.00661062970502241174d0, 0.00909530922354827295d0, -0.20160072778491013140d0, &
         0.51169696718727644908d0, 0.00000000003147682272d0, -0.00000000048465972408d0, &
         0.00000000063675740242d0, 0.00000003377623323271d0, -0.00000015451139637086d0, &
         -0.00000203340624738438d0, 0.00001947204525295057d0, 0.00002854147231653228d0, &
         -0.00101565063152200272d0, 0.00271187003520095655d0, 0.02328095035422810727d0, &
         -0.16725021123116877197d0, 0.32490054966649436974d0, 0.00000000002319363370d0, &
         -0.00000000006303206648d0, -0.00000000264888267434d0, 0.00000002050708040581d0, &
         0.00000011371857327578d0, -0.00000211211337219663d0, 0.00000368797328322935d0, &
         0.00009823686253424796d0, -0.00065860243990455368d0, -0.00075285814895230877d0, &
         0.02585434424202960464d0, -0.11637092784486193258d0, 0.18267336775296612024d0, &
         -0.00000000000367789363d0, 0.00000000020876046746d0, -0.00000000193319027226d0, &
         -0.00000000435953392472d0, 0.00000018006992266137d0, -0.00000078441223763969d0, &
         -0.00000675407647949153d0, 0.00008428418334440096d0, -0.00017604388937031815d0, &
         -0.00239729611435071610d0, 0.02064129023876022970d0, -0.06905562880005864105d0, &
         0.09084526782065478489d0 /)
    real(fdp) :: w, t, y
    integer :: k

    w = abs(x)

    if (w < 2.2d0) then
       t = w * w
       k = int(t)
       t = t - dble(k)
       k = k * 13
       y = ((((((((((((a(k) * t + a(k + 1)) * t + &
            a(k + 2)) * t + a(k + 3)) * t + a(k + 4)) * t + &
            a(k + 5)) * t + a(k + 6)) * t + a(k + 7)) * t + &
            a(k + 8)) * t + a(k + 9)) * t + a(k + 10)) * t + &
            a(k + 11)) * t + a(k + 12)) * w
    else if (w < 6.9d0) then
       k = int(w)
       t = w - dble(k)
       k = 13 * (k - 2)
       y = (((((((((((b(k) * t + b(k + 1)) * t + &
            b(k + 2)) * t + b(k + 3)) * t + b(k + 4)) * t + &
            b(k + 5)) * t + b(k + 6)) * t + b(k + 7)) * t + &
            b(k + 8)) * t + b(k + 9)) * t + b(k + 10)) * t + &
            b(k + 11)) * t + b(k + 12)
       y = y * y
       y = y * y
       y = y * y
       y = 1.0d0 - y * y
    else
       y = 1.0d0
    end if

    f = sign(y, x)
  end function f_erf



  recursive function f_ierfc(x) result(f)
    !
    !  Inverse of the complimentary error function
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp) :: y, z, u, s, t, w
    real(fdp), parameter :: qa = 9.16461398268964d-1, qb =&
         & 2.31729200323405d-1, qc = 4.88826640273108d-1, qd =&
         & 1.24610454613712d-1, q0 = 4.99999303439796d-1, q1 =&
         & 1.16065025341614d-1, q2 = 1.50689047360223d-1, q3 =&
         & 2.69999308670029d-1, q4 = -7.28846765585675d-2, pa =&
         & 3.97886080735226000d0, pb = 1.20782237635245222d-1, p0 =&
         & 2.44044510593190935d-1, p1 = 4.34397492331430115d-1, p2 =&
         & 6.86265948274097816d-1, p3 = 9.56464974744799006d-1, p4 =&
         & 1.16374581931560831d0, p5 = 1.21448730779995237d0, p6 =&
         & 1.05375024970847138d0, p7 = 7.13657635868730364d-1, p8 =&
         & 3.16847638520135944d-1, p9 = 1.47297938331485121d-2, p10 =&
         & -1.05872177941595488d-1, p11 = -7.43424357241784861d-2,&
         & p12 = 2.20995927012179067d-3, p13 = 3.46494207789099922d-2&
         &, p14 = 1.42961988697898018d-2, p15 = -1.18598117047771104d&
         &-2, p16 = -1.12749169332504870d-2, p17 =&
         & 3.39721910367775861d-3, p18 = 6.85649426074558612d-3, p19 &
         &= -7.71708358954120939d-4, p20 = -3.51287146129100025d-3,&
         & p21 = 1.05739299623423047d-4, p22 = 1.12648096188977922d-3

    if (x == 0.0d0) then
       f = inf
    else if (x == 1.0d0) then
       f = 0.0d0
    else
       z = x
       
       if (x  >  1.0d0) z = 2.0d0 - x
       
       w = qa - log(z)
       u = sqrt(w)
       s = (qc + log(u)) / w
       t = 1.0d0 / (u + qb)
       y = u * (1.0d0 - s * (0.5d0 + s * qd)) - ((((q4 * t + q3) * t + q2) *&
            & t + q1) * t + q0) * t
       t = pa / (pa + y)
       u = t - 0.5d0
       s = (((((((((p22 * u + p21) * u + p20) * u + p19) * u + p18) * u &
            &+ p17) * u + p16) * u + p15) * u + p14) * u + p13) * u + p12
       s = ((((((((((((s * u + p11) * u + p10) * u + p9) * u + p8) * u +&
            & p7) * u + p6) * u + p5) * u + p4) * u + p3) * u + p2) * u &
            &+ p1) * u + p0) * t - z * exp(y * y - pb)
       y = y + s * (1.0d0 + y * s)
       
       if (x  >  1.0d0) y = -y
       
       f = y
    end if
  end function f_ierfc



  recursive function f_gamma(x) result(f)
    !
    ! Gamma function in double precision
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp) :: w, y
    integer :: n, k
    real(fdp), parameter :: p0 = 0.999999999999999990d0, p1 = &
         &-0.422784335098466784d0, p2 = -0.233093736421782878d0, p3 =&
         & 0.191091101387638410d0, p4 = -0.024552490005641278d0, p5 =&
         & -0.017645244547851414d0, p6 = 0.008023273027855346d0, p7 =&
         & -0.000804329819255744d0, p8 = -0.000360837876648255d0, p9 &
         &= 0.000145596568617526d0, p10 = -0.000017545539395205d0,&
         & p11 = -0.000002591225267689d0, p12 =&
         & 0.000001337767384067d0, p13 = -0.000000199542863674d0

      n = nint(x - 2)
      w = x - (n + 2)
      y = ((((((((((((p13 * w + p12) * w + p11) * w + p10) * w + p9) &
           &* w + p8) * w + p7) * w + p6) * w + p5) * w + p4) * w +&
           & p3) * w + p2) * w + p1) * w + p0
      if (n  >  0) then
          w = x - 1
          do k = 2, n
              w = w * (x - k)
          end do
      else
          w = 1
          do k = 0, -n - 1
              y = y * (x + k)
          end do
      end if

      f = w / y
  end function f_gamma



  recursive function f_lgamma(x) result(f)
    !
    ! log gamma function
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp) :: v, w, y, t
    integer :: k
    real(fdp), dimension(0:21), parameter :: a = (/ 0.00009967270908702825d0, &
         -0.00019831672170162227d0, -0.00117085315349625822d0, 0.00722012810948319552d0, &
         -0.00962213009367802970d0, -0.04219772092994235254d0,0.16653861065243609743d0, & 
         -0.04200263501129018037d0, -0.65587807152061930091d0, 0.57721566490153514421d0, &
         0.99999999999999999764d0, 0.00004672097259011420d0, -0.00006812300803992063d0, &
         -0.00132531159076610073d0, 0.00733521178107202770d0, -0.00968095666383935949d0, &
         -0.04217642811873540280d0, 0.16653313644244428256d0, -0.04200165481709274859d0, &
         -0.65587818792782740945d0, 0.57721567315209190522d0, 0.99999999973565236061d0 /)
    real(fdp), dimension(0:97), parameter :: b = (/ -0.00000000004587497028d0, &
         0.00000000019023633960d0, 0.00000000086377323367d0, 0.00000001155136788610d0, &
         -0.00000002556403058605d0, -0.00000015236723372486d0, -0.00000316805106385740d0, &
         0.00000122903704923381d0, 0.00002334372474572637d0, 0.00111544038088797696d0, &
         0.00344717051723468982d0, 0.03198287045148788384d0, -0.32705333652955399526d0, &
         0.40120442440953927615d0, -0.00000000005184290387d0, -0.00000000083355121068d0, &
         -0.00000000256167239813d0, 0.00000001455875381397d0, 0.00000013512178394703d0, &
         0.00000029898826810905d0, -0.00000358107254612779d0, -0.00002445260816156224d0, &
         -0.00004417127762011821d0, 0.00112859455189416567d0, 0.00804694454346728197d0, &
         0.04919775747126691372d0, -0.24818372840948854178d0, 0.11071780856646862561d0, &
         0.00000000030279161576d0, 0.00000000160742167357d0, -0.00000000405596009522d0, &
         -0.00000005089259920266d0, -0.00000002029496209743d0, 0.00000135130272477793d0, &
         0.00000391430041115376d0, -0.00002871505678061895d0, -0.00023052137536922035d0, &
         0.00045534656385400747d0, 0.01153444585593040046d0, 0.07924014651650476036d0, &
         -0.12152192626936502982d0, -0.07916438300260539592d0, -0.00000000050919149580d0, &
         -0.00000000115274986907d0, 0.00000001237873512188d0, 0.00000002937383549209d0, &
         -0.00000030621450667958d0, -0.00000077409414949954d0, 0.00000816753874325579d0, &
         0.00002412433382517375d0, -0.00026061217606063700d0, -0.00091000087658659231d0, &
         0.01068093850598380797d0, 0.11395654404408482305d0, 0.07209569059984075595d0, &
         -0.10971041451764266684d0, 0.00000000040119897187d0, -0.00000000013224526679d0, &
         -0.00000001002723190355d0, 0.00000002569249716518d0, 0.00000020336011868466d0, &
         -0.00000118097682726060d0, -0.00000300660303810663d0, 0.00004402212897757763d0, &
         -0.00001462405876235375d0, -0.00164873795596001280d0, 0.00513927520866443706d0, &
         0.13843580753590579416d0, 0.32730190978254056722d0, 0.08588339725978624973d0, &
         -0.00000000015413428348d0, 0.00000000064905779353d0, 0.00000000160702811151d0, &
         -0.00000002655645793815d0, 0.00000007619544277956d0, 0.00000047604380765353d0, &
         -0.00000490748870866195d0, 0.00000821513040821212d0, 0.00014804944070262948d0, &
         -0.00122152255762163238d0, -0.00087425289205498532d0, 0.14438703699657968310d0, &
         0.61315889733595543766d0, 0.55513708159976477557d0, 0.00000000001049740243d0, &
         -0.00000000025832017855d0, 0.00000000139591845075d0, -0.00000000021177278325d0, &
         -0.00000005082950464905d0, 0.00000037801785193343d0, -0.00000073982266659145d0, &
         -0.00001088918441519888d0, 0.00012491810452478905d0, -0.00049171790705139895d0, &
         -0.00425707089448266460d0, 0.13595080378472757216d0, 0.89518356003149514744d0, &
         1.31073912535196238583d0 /)
    real(fdp), dimension(0:64), parameter :: c = (/ 0.0000000116333640008d0, &
         -0.0000000833156123568d0, 0.0000003832869977018d0, -0.0000015814047847688d0, &
         0.0000065010672324100d0, -0.0000274514060128677d0, 0.0001209015360925566d0, &
         -0.0005666333178228163d0, 0.0029294103665559733d0, -0.0180340086069185819d0, &
         0.1651788780501166204d0, 1.1031566406452431944d0, 1.2009736023470742248d0, &
         0.0000000013842760642d0, -0.0000000069417501176d0, 0.0000000342976459827d0, &
         -0.0000001785317236779d0, 0.0000009525947257118d0, -0.0000052483007560905d0, &
         0.0000302364659535708d0, -0.0001858396115473822d0, 0.0012634378559425382d0, &
         -0.0102594702201954322d0, 0.1243625515195050218d0, 1.3888709263595291174d0, &
         2.4537365708424422209d0, 0.0000000001298977078d0, -0.0000000008029574890d0, &
         0.0000000049454846150d0, -0.0000000317563534834d0, 0.0000002092136698089d0, &
         -0.0000014252023958462d0, 0.0000101652510114008d0, -0.0000774550502862323d0, &
         0.0006537746948291078d0, -0.0066014912535521830d0, 0.0996711934948138193d0, &
         1.6110931485817511402d0, 3.9578139676187162939d0, 0.0000000000183995642d0, &
         -0.0000000001353537034d0, 0.0000000009984676809d0, -0.0000000076346363974d0, &
         0.0000000599311464148d0, -0.0000004868554120177d0, 0.0000041441957716669d0, &
         -0.0000377160856623282d0, 0.0003805693126824884d0, -0.0045979851178130194d0, &
         0.0831422678749791178d0, 1.7929113303999329439d0, 5.6625620598571415285d0, &
         0.0000000000034858778d0, -0.0000000000297587783d0, 0.0000000002557677575d0, &
         -0.0000000022705728282d0, 0.0000000207024992450d0, -0.0000001954426390917d0, &
         0.0000019343161886722d0, -0.0000204790249102570d0, 0.0002405181940241215d0, &
         -0.0033842087561074799d0, 0.0713079483483518997d0, 1.9467574842460867884d0, &
         7.5343642367587329552d0 /)
    real(fdp), dimension(0:6), parameter :: d = (/ -0.00163312359200500807d0, &
         0.00083644533703385956d0, -0.00059518947575728181d0, 0.00079365057505415415d0, &
         -0.00277777777735463043d0, 0.08333333333333309869d0, 0.91893853320467274178d0 /)

    w = merge(1-x, x, x < 0.0d0)

    if (w < 0.5d0) then
       k = merge(11, 0, w >= 0.25d0)
       y = ((((((((((a(k) * w + a(k + 1)) * w + &
            a(k + 2)) * w + a(k + 3)) * w + a(k + 4)) * w + &
            a(k + 5)) * w + a(k + 6)) * w + a(k + 7)) * w + &
            a(k + 8)) * w + a(k + 9)) * w + a(k + 10)) * w
       y = -log(y)
    else if (w < 3.5d0) then
       t = w - 4.5d0 / (w + 0.5d0)
       k = int(t + 4.0d0)
       t = t - (dble(k) - 3.5d0)
       k = k * 14
       y = ((((((((((((b(k) * t + b(k + 1)) * t + &
            b(k + 2)) * t + b(k + 3)) * t + b(k + 4)) * t + &
            b(k + 5)) * t + b(k + 6)) * t + b(k + 7)) * t + &
            b(k + 8)) * t + b(k + 9)) * t + b(k + 10)) * t + &
            b(k + 11)) * t + b(k + 12)) * t + b(k + 13)
    else if (w < 8.0d0) then
       k = int(w) - 3
       t = w - (dble(k) + 3.5d0)
       k = k * 13
       y = (((((((((((c(k) * t + c(k + 1)) * t + &
            c(k + 2)) * t + c(k + 3)) * t + c(k + 4)) * t + &
            c(k + 5)) * t + c(k + 6)) * t + c(k + 7)) * t + &
            c(k + 8)) * t + c(k + 9)) * t + c(k + 10)) * t + &
            c(k + 11)) * t + c(k + 12)
    else
       v = 1.0d0 / w
       t = v * v
       y = (((((d(0) * t + d(1)) * t + d(2)) * t + &
            d(3)) * t + d(4)) * t + d(5)) * v + d(6)
       y = y + ((w - 0.5d0) * log(w) - w)
    end if

    f = merge(log(pi / sin(pi * x)) - y, y, x < 0.0d0)
  end function f_lgamma



  recursive function f_cbrt(x) result(f)
    !
    !  Cubic root of a number
    !

    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp), dimension(0:23), parameter :: c = (/ 1.5319394088521d&
         &-3, -1.8843445653409d-2, 1.0170534986000d-1, &
         &-3.1702448761286d-1, 6.3520892642253d-1, -8.8106985991189d&
         &-1, 1.0517503764540d0, 4.2674123235580d-1, 1.5079083659190d&
         &-5, -3.7095709111375d-4, 4.0043972242353d-3, &
         &-2.4964114079723d-2, 1.0003913718511d-1, -2.7751961573273d&
         &-1, 6.6256121926465d-1, 5.3766026150315d-1,&
         & 1.4842542902609d-7, -7.3027601203435d-6, 1.5766326109233d&
         &-4, -1.9658008013138d-3, 1.5755176844105d-2, &
         &-8.7413201405100d-2, 4.1738741349777d-1, 6.7740948115980d-1&
         & /)
    real(fdp), parameter :: p2pow16 = 65536.0d0, p2pow48 =&
         & 281474976710656.0d0, p2powm16 = 1 / p2pow16, p2powm48 = 1 &
         &/ p2pow48
    real(fdp) :: w, y, u
    integer :: k

    if (x .eq. 0) then
       f = 0
       return
    end if

    w = abs(x)
    y = sign(0.5d0, x)

    if (w  >  8) then
       do while (w  >  p2pow48)
          w = w * p2powm48
          y = y * p2pow16
       end do

       do while (w  >  8)
          w = w * 0.125d0
          y = y * 2
       end do
    else if (w  <  1) then
       do while (w  <  p2powm48)
          w = w * p2pow48
          y = y * p2powm16
       end do

       do while (w  <  1)
          w = w * 8
          y = y * 0.5d0
       end do
    end if

    if (w  <  2) then
       k = 0
    else if (w  <  4) then
       k = 8
    else
       k = 16
    end if

    u = ((((((c(k) * w + c(k + 1)) * w + c(k + 2)) * w + c(k + 3)) *&
         & w + c(k + 4)) * w + c(k + 5)) * w + c(k + 6)) * w + c(k +&
         & 7)
    f = y * (u + 3 * u * w / (w + 2 * u * u * u))
  end function f_cbrt



  recursive function f_fresc(x) result(f)
    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp) :: xa, px, t, t0, t1, t2, r, a, a0, a1, g, s, su, q
    integer :: k, m

    xa=abs(x)
    px=pi*xa
    t=.5d0*px*xa
    t2=t*t

    if (xa == 0.0d0) then
       f=0.0d0
    else if (xa < 2.5d0) then
       r=xa
       f=r
       do k=1,50
          r=-.5d0*r*(4.0d0*k-3.0d0)/k/(2.0d0*k-1.0d0)/(4.0d0*k+1.0d0)*t2
          f=f+r
          if (abs(r) < abs(f)*eps) exit
       end do
    else if (xa < 4.5d0) then
       m=int(42.0+1.75*t)
       su=0.0d0
       f=0.0d0
       s=0.0d0
       a1=0.0d0
       a0=1.0d-100
       do k=m,0,-1
          a=(2.0d0*k+3.0d0)*a0/t-a1
          if (k == int(k/2)*2) then
             f=f+a
          endif
          su=su+(2.0d0*k+1.0d0)*a*a
          a1=a0
          a0=a
       end do
       q=sqrt(su)
       f=f*xa/q
    else
       r=1.0d0
       a=1.0d0
       do k=1,20
          r=-.25d0*r*(4.0d0*k-1.0d0)*(4.0d0*k-3.0d0)/t2
          a=a+r
       end do
       r=1.0d0/(px*xa)
       g=r
       do  k=1,12
          r=-.25d0*r*(4.0d0*k+1.0d0)*(4.0d0*k-1.0d0)/t2
          g=g+r
       end do
       t0=t-int(t/(2.0d0*pi))*2.0d0*pi
       f=.5d0+(a*sin(t0)-g*cos(t0))/px
    endif

    if (x < 0.0d0) f=-f
  end function f_fresc



  recursive function f_fress(x) result(f)
    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp) :: xa, px, t, t0, t1, t2, r, a, a0, a1, g, s, su, q
    integer :: k, m

    xa=abs(x)
    px=pi*xa
    t=.5d0*px*xa
    t2=t*t

    if (xa == 0.0d0) then
       f=0.0d0
    else if (xa < 2.5d0) then
       f=xa*t/3.0d0
       r=f
       do k=1,50
          r=-.5d0*r*(4.0d0*k-1.0d0)/k/(2.0d0*k+1.0d0)/(4.0d0*k+3.0d0)*t2
          f=f+r
          if (abs(r) < abs(f)*eps) exit
       end do
    else if (xa < 4.5d0) then
       m=int(42.0+1.75*t)
       su=0.0d0
       f=0.0d0
       a1=0.0d0
       a0=1.0d-100
       do k=m,0,-1
          a=(2.0d0*k+3.0d0)*a0/t-a1
          if (k /= int(k/2)*2) then
             f=f+a
          endif
          su=su+(2.0d0*k+1.0d0)*a*a
          a1=a0
          a0=a
       end do
       q=sqrt(su)
       f=f*xa/q
    else
       r=1.0d0
       a=1.0d0
       do k=1,20
          r=-.25d0*r*(4.0d0*k-1.0d0)*(4.0d0*k-3.0d0)/t2
          a=a+r
       end do
       r=1.0d0/(px*xa)
       g=r
       do  k=1,12
          r=-.25d0*r*(4.0d0*k+1.0d0)*(4.0d0*k-1.0d0)/t2
          g=g+r
       end do
       t0=t-int(t/(2.0d0*pi))*2.0d0*pi
       f=.5d0-(a*cos(t0)+g*sin(t0))/px
    endif

    if (x < 0.0d0) f=-f
  end function f_fress



  recursive function f_expi(x) result(f)
    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp) :: r
    integer :: k, m

    if (x < -1.0d0) then
       m=20-int(80.0/x)
       r=0.0d0
       do k=m,1,-1
          r=k/(1.0d0+dble(k)/(r-x))
       end do
       f=-exp(x)*1.0d0/(r-x)
    else if (x < 0.0d0) then
       f=1.0d0
       r=1.0d0
       do k=1,25
          r=r*dble(k)*x/(dble(k)+1.0d0)**2
          f=f+r
          if (abs(r) <= abs(f)*eps) exit
       end do
       f=0.5772156649015328d0+log(-x)+x*f
    else if (x == 0.0d0) then
       f = -inf
    else if (x <= 40.0d0) then
       f=1.0d0
       r=1.0d0
       do k=1,100
          r=r*dble(k)*x/(dble(k)+1.0d0)**2
          f=f+r
          if (abs(r/f) <= eps) exit
       end do
       f=0.5772156649015328d0+log(x)+x*f
    else
       f=1.0d0
       r=1.0d0
       do k=1,20
          r=r*dble(k)/x
          f=f+r
       end do
       f=exp(x)/x*f
    endif
  end function f_expi



  recursive function f_sini(x) result(f)
    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp), dimension(101) :: bj
    real(fdp) :: xr, xa1, xa0, xa, xgl, xs, xcs, xss, xg, xg1, xg2, xf
    integer :: k, m

    if (x == 0.0d0) then
       f=0.0d0
    else if (x <= 16.0d0) then
       xr=x
       f=x
       do k=1,40
          xr=-.5d0*xr*(2*k-1)/k/(4*k*k+4*k+1)*x**2
          f=f+xr
          if (abs(xr) < abs(f)*eps) return
       end do
    else if (x <= 32.0d0) then
       m=int(47.2+.82*x)
       xa1=0.0d0
       xa0=1.0d-100
       do k=m,1,-1
          xa=4.0d0*k*xa0/x-xa1
          bj(k)=xa
          xa1=xa0
          xa0=xa
       end do
       xs=bj(1)
       do k=3,m,2
          xs=xs+2.0d0*bj(k)
       end do
       bj(1)=bj(1)/xs
       do k=2,m
          bj(k)=bj(k)/xs
       end do
       xr=1.0d0
       xg1=bj(1)
       do k=2,m
          xr=.25d0*xr*(2.0*k-3.0)**2/((k-1.0)*(2.0*k-1.0)**2)*x
          xg1=xg1+bj(k)*xr
       end do
       xr=1.0d0
       xg2=bj(1)
       do k=2,m
          xr=.25d0*xr*(2.0*k-5.0)**2/((k-1.0)*(2.0*k-3.0)**2)*x
          xg2=xg2+bj(k)*xr
       end do
       xcs=cos(x/2.0d0)
       xss=sin(x/2.0d0)
       f=x*xcs*xg1+2*xss*xg2-sin(x)
    else
       xr=1.0d0
       xf=1.0d0
       do k=1,9
          xr=-2.0d0*xr*k*(2*k-1)/x**2
          xf=xf+xr
       end do
       xr=1.0d0/x
       xg=xr
       do k=1,8
          xr=-2.0d0*xr*(2*k+1)*k/x**2
          xg=xg+xr
       end do
       f=pi2-xf*cos(x)/x-xg*sin(x)/x
    endif
  end function f_sini



  recursive function f_cosi(x) result(f)
    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp), dimension(101) :: bj
    real(fdp) :: xr, xa1, xa0, xa, xgl, xs, xcs, xss, xg, xg1, xg2, xf
    integer :: k, m

    if (x == 0.0d0) then
       f = -inf
    else if (x <= 16.0d0) then
       xr=-.25d0*x**2
       f=el+log(x)+xr
       do k=2,40
          xr=-.5d0*xr*(k-1)/(k*k*(2*k-1))*x**2
          f=f+xr
          if (abs(xr) < abs(f)*eps) exit
       end do
    else if (x <= 32.0d0) then
       m=int(47.2+.82*x)
       xa1=0.0d0
       xa0=1.0d-100
       do k=m,1,-1
          xa=4.0d0*k*xa0/x-xa1
          bj(k)=xa
          xa1=xa0
          xa0=xa
       end do
       xs=bj(1)
       do k=3,m,2
          xs=xs+2.0d0*bj(k)
       end do
       bj(1)=bj(1)/xs
       do k=2,m
          bj(k)=bj(k)/xs
       end do
       xr=1.0d0
       xg1=bj(1)
       do k=2,m
          xr=.25d0*xr*(2.0*k-3.0)**2/((k-1.0)*(2.0*k-1.0)**2)*x
          xg1=xg1+bj(k)*xr
       end do
       xr=1.0d0
       xg2=bj(1)
       do k=2,m
          xr=.25d0*xr*(2.0*k-5.0)**2/((k-1.0)*(2.0*k-3.0)**2)*x
          xg2=xg2+bj(k)*xr
       end do
       xcs=cos(x/2.0d0)
       xss=sin(x/2.0d0)
       f=el+log(x)-x*xss*xg1+2*xcs*xg2-2*xcs*xcs
    else
       xr=1.0d0
       xf=1.0d0
       do k=1,9
          xr=-2.0d0*xr*k*(2*k-1)/x**2
          xf=xf+xr
       end do
       xr=1.0d0/x
       xg=xr
       do k=1,8
          xr=-2.0d0*xr*(2*k+1)*k/x**2
          xg=xg+xr
       end do
       f=xf*sin(x)/x-xg*cos(x)/x
    endif
  end function f_cosi



  recursive function f_logi(x) result(f)
    real(fdp), intent(in) :: x
    real(fdp) :: f

    if (x == 0.0d0) then
       f = 0.0d0
    else if (x == 1.0d0) then
       f = -inf
    else if (x == 1.451369234883381050282962414152657d0) then
       f = 0.0d0
    else if (x == 2.0d0) then
       f = 1.0451637801174927848d0
    else
       f = f_expi(log(x))
    end if
  end function f_logi



  recursive function f_ierf(x) result(f)
    real(fdp), intent(in) :: x
    real(fdp) :: f

    if (x == 0.0d0) then
       f = 0.0d0
    else if (x == 1) then
       f = inf
    else
       f = f_ierfc(1-x)
    end if
  end function f_ierf



  recursive function f_elle(x) result(f)
    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp) :: ae, be, pk

    if (x == 1.0) then
       f=1.0d0
    else if (x == 0.0d0) then
       f = pi2
    else if (x == -1.0d0) then
       f = 1.910098894513856d0
    else
       pk=1.0d0-x*x
       ae=(((.01736506451d0*pk+.04757383546d0)*pk+.0626060122d0)*pk+.44325141463d0)*pk+1.0d0
       be=(((.00526449639d0*pk+.04069697526d0)*pk+.09200180037d0)*pk+.2499836831d0)*pk
       f=ae-be*log(pk)
    endif
  end function f_elle



  recursive function f_ellk(x) result(f)
    real(fdp), intent(in) :: x
    real(fdp) :: f

    real(fdp) :: pk, ak, bk

    if (x == 1.0) then
       f = inf
    else if (x == 0.0d0) then
       f = pi2
    else if (x == -1.0d0) then
       f = 1.311028777146060d0
    else if (x == 0.5d0) then
       f = 1.685750354812596d0
    else
       pk=1.0d0-x*x
       ak=(((.01451196212d0*pk+.03742563713d0)*pk+.03590092383d0)*pk+.09666344259d0)*pk+1.38629436112d0
       bk=(((.00441787012d0*pk+.03328355346d0)*pk+.06880248576d0)*pk+.12498593597d0)*pk+.5d0
       f=ak-bk*log(pk)
    endif
  end function f_ellk



  recursive function f_ielle(x, phi) result(f)
    real(fdp), intent(in) :: x, phi
    real(fdp) :: f

    real(fdp) :: g, a, a0, b, b0, c, r, d, d0, ce, ck, fac
    integer :: n

    if (phi == 0.0d0) then
       f = 0.0d0
    else if (x == 1.0d0 .and. phi == pi2) then
       f = 1.0d0
    else if (x == 1.0d0) then
       f = sin(phi)
    else if (x == 0.0d0) then
       f = phi
    else
       g=0.0d0
       a0=1.0d0
       b0=sqrt(1.0d0-x*x)
       d0 = phi
       r=x*x
       fac=1.0d0

       do n=1,40
          a=(a0+b0)/2.0d0
          b=sqrt(a0*b0)
          c=(a0-b0)/2.0d0
          fac=2.0d0*fac
          r=r+fac*c*c
          if (phi /= pi2) then
             d=d0+atan((b0/a0)*tan(d0))
             g=g+c*sin(d)
             d0=d+pi*int(d/pi+.5d0)
          endif
          a0=a
          b0=b
          if (c < 1.0d-7) exit
       end do
       ce=pi*(2.0d0-r)/(4.0d0*a)
       if (phi == pi2) then
          f = ce
       else
          ck=pi/(2.0d0*a)
          f =d/fac/a*ce/ck+g
       endif
    endif
  end function f_ielle



  recursive function f_iellf(x, phi) result(f)
    real(fdp), intent(in) :: x, phi
    real(fdp) :: f

    real(fdp) :: g, a, a0, b, b0, c, r, d, d0, ck, fac
    integer :: n

    if (phi == 0) then
       f = 0.0d0
    else if (x == 0.0d0) then
       f = x
    else if (x == 1.0d0 .and. phi == pi2) then
       f = inf
    else if (x == 1.0d0) then
       f = log(1.0d0+sin(phi))/cos(phi)
    else
       g=0.0d0
       a0=1.0d0
       b0=sqrt(1.0d0-x*x)
       d0 = phi
       r=x*x
       fac=1.0d0

       do n=1,40
          a=(a0+b0)/2.0d0
          b=sqrt(a0*b0)
          c=(a0-b0)/2.0d0
          fac=2.0d0*fac
          r=r+fac*c*c
          if (phi /= pi2) then
             d=d0+atan((b0/a0)*tan(d0))
             g=g+c*sin(d)
             d0=d+pi*int(d/pi+.5d0)
          endif
          a0=a
          b0=b
          if (c < 1.0d-7) exit
       end do

       if (phi == pi2) then
          ck=pi/(2.0d0*a)
          f=ck
       else
          f=d/(fac*a)
       endif
    endif
  end function f_iellf

end module extrafunc

module interpreter
  use params
  use extrafunc

  implicit none

  private
  public :: s_createfn, s_errprecision, s_evaluatefn, s_evaluateerr, s_destroyfn, f_numbervars

  type cachefn
     real(fdp), dimension(:), pointer :: numbers
     integer, dimension(:), pointer :: commands
  end type cachefn

  character(len=variablelength), dimension(:), allocatable :: variables
  integer, dimension(:), allocatable :: lenvariables
  integer :: numvariables
  real(fdp), dimension(:), allocatable :: varerrors
  character(len=errorlength) :: error
  type(cachefn), dimension(:), pointer, save :: cachedfuncs

contains

  subroutine s_firstrun()
    logical, save :: firstrun = .true.
    integer :: i

    if (firstrun) then
       allocate(cachedfuncs(10))

       do i=1, 10
          nullify(cachedfuncs(i)%numbers, cachedfuncs(i)%commands)
       end do

       firstrun = .false.
    end if
  end subroutine s_firstrun


  
  subroutine s_seterror(t_error)
    character(len=*), intent(in) :: t_error

    if (error == 'OK' .and. len_trim(t_error) > 0) error = trim(t_error)
  end subroutine s_seterror



  function f_clean(func) result(tmpfunc)
    character(len=*), intent(in) :: func
    character(len=len(func)) :: tmpfunc
    integer :: i, c, l
    
    tmpfunc = adjustl(func)
    l = len_trim(tmpfunc)
    i = 1

    do while (i <= l)
       if (tmpfunc(i:i)== ' ') then
          tmpfunc(i:) = tmpfunc(i+1:)
          l = l - 1
       else
          c = ichar(tmpfunc(i:i))
          if ((c <= ichar('Z')) .and. (c >= ichar('A'))) tmpfunc(i:i) = achar(c-ichar('A')+ichar('a'))
          i = i + 1
       end if
    end do
  end function f_clean



  function f_numbervars(vars) result(numvars)
    character(len=*), intent(in) :: vars
    integer :: numvars
    character(len=len(vars)) :: tmpvars
    character(len=variablelength) :: tmpvar

    tmpvars = trim(adjustl(vars))
    numvars = 0

    do while (len_trim(tmpvars) > 0)
       read(tmpvars, *) tmpvar
       numvars = numvars + 1
       tmpvars = tmpvars(index(tmpvars, trim(tmpvar))+len_trim(tmpvar):)
    end do
  end function f_numbervars



  subroutine s_errprecision(number)
    real(fdp), intent(inout) :: number
    real(fdp) :: tmp1, tmp2

    !
    ! Format the error to the correct precision
    !

    if (number /= 0.0d0) then
       tmp1 = 10.0d0**(floor(log10(number)))
       tmp2 = anint(number/tmp1)

       if (tmp2 > 1.0d0) then
          number = tmp2*tmp1
       else
          tmp1 = 10.0d0**(floor(log10(number))-1.0d0)
          tmp2 = anint(number/tmp1)
          number = tmp2*tmp1
       end if
    end if
  end subroutine s_errprecision



  subroutine s_readvarerror(errorstring, errorval)
    character(len=*), intent(in) :: errorstring
    real(fdp) :: errorval
    integer :: command, lenerror
    character(len=errorlength) :: tmperror

    errorval = 0.0d0
    lenerror = len_trim(errorstring)
    command = nop_

    if (errorstring(1:1) == '[' .and. errorstring(lenerror:lenerror) == ']') then
       tmperror = errorstring(2:lenerror-1)
       
       call s_readnumber(tmperror, lenerror, command, errorval)

       if (command /= number_) call s_seterror('Not a number: '//tmperror)

       if (len_trim(tmperror) /= 0) call s_seterror('Not correct error format: '//tmperror)
    else
       call s_seterror('Not a correct error: '//errorstring)
    end if
  end subroutine s_readvarerror



  subroutine s_allocatevariables(vars)
    character(len=*), intent(in) :: vars
    integer :: numvars, i, bracketpos, j, c
    character(len=variablelength+errorlength), dimension(:), allocatable :: tmp

    numvars = f_numbervars(vars)

    allocate(tmp(numvars), variables(numvars), lenvariables(numvars), varerrors(numvars))

    read(vars, *) tmp

    do i = 1, numvars
       do j = 1, len_trim(tmp(i))
          c = ichar(tmp(i)(j:j))
          if ((c <= ichar('Z')) .and. (c >= ichar('A'))) tmp(i)(j:j) = achar(c-ichar('A')+ichar('a'))
       end do

       bracketpos = scan(tmp(i), "[")

       if (bracketpos /= 0) then
          variables(i) = trim(tmp(i)(:bracketpos-1))

          call s_readvarerror(trim(tmp(i)(bracketpos:)), varerrors(i))
       else
          variables(i) = trim(tmp(i))
          varerrors(i) = 0.0d0
       end if
    end do

    lenvariables = len_trim(variables)
    numvariables = numvars

    deallocate(tmp)
  end subroutine s_allocatevariables



  subroutine s_deallocatevariables()
    deallocate(variables, lenvariables, varerrors)
  end subroutine s_deallocatevariables



  subroutine s_funcpush(funcstack, funcstackn, command, number)
    type(cachefn), intent(inout) :: funcstack
    integer, intent(inout) :: funcstackn
    real(fdp), intent(in) :: number
    integer, intent(in) :: command
    type(cachefn) :: tmpstack

    funcstackn = funcstackn + 1

    if (funcstackn >= size(funcstack%numbers)) then
       allocate(tmpstack%numbers(funcstackn+5), tmpstack%commands(funcstackn+5))
       tmpstack%numbers(:funcstackn-1) = funcstack%numbers
       tmpstack%commands(:funcstackn-1) = funcstack%commands
       tmpstack%numbers(funcstackn:) = 0.0d0
       tmpstack%commands(funcstackn:) = end_
       deallocate(funcstack%numbers, funcstack%commands)
       funcstack%numbers => tmpstack%numbers
       funcstack%commands => tmpstack%commands
    end if

    funcstack%commands(funcstackn) = command
    funcstack%numbers(funcstackn) = number
  end subroutine s_funcpush



  subroutine s_checkoperations(func, lenfunc, command, nextcommand)
    integer, intent(in) :: lenfunc
    character(lenfunc), intent(in) :: func
    integer, intent(in) :: command, nextcommand

    select case(command)

    case(number_)
       select case(nextcommand)

       case(number_)
          call s_seterror('Number following number: '//func)

       case(variablestart:variableend)
          call s_seterror('Variable following number: '//func)

       case(bracket_)
          call s_seterror('Bracket following number: '//func)

       case(expressionstart:expressionend)
          call s_seterror('Expression following number: '//func)

       end select
       
    case(variablestart:variableend)
       select case(nextcommand)
          
       case(number_)
          call s_seterror('Number following variable: '//func)
          
       case(variablestart:variableend)
          call s_seterror('Variable following variable: '//func)
          
       case(bracket_)
          call s_seterror('Bracket following variable: '//func)
          
       case(expressionstart:expressionend)
          call s_seterror('Expression following variable: '//func)
          
       end select
       
    case(conditionalstart:conditionalend)
       select case(nextcommand)

       case(conditionalstart:conditionalend, opsstart:opsend)
          call s_seterror('Operator following operator: '//func)

       case(end_)
          call s_seterror('Premature end of function')

       end select

    case(opsstart:opsend, pmstart:pmend)
       select case(nextcommand)

       case(operatorstart:operatorend)
          call s_seterror('Operator following operator: '//func)

       case(end_)
          call s_seterror('Premature end of function')

       end select
    end select
  end subroutine s_checkoperations



  subroutine s_checkcomponents(components, command)
    integer, intent(in) :: components, command
    
    if (components /= commandcomps(command)) call s_seterror('Incorrect components for: '//expressions(command))
  end subroutine s_checkcomponents



  subroutine s_readnumber(func, lenfunc, command, number)
    integer, intent(inout) :: lenfunc
    character(lenfunc), intent(inout) :: func
    integer, intent(out) :: command
    real(fdp), intent(out) :: number
    integer :: pos
    logical :: dflag, exponent, finish

    exponent = .false.
    dflag = .false.
    finish = .false.
    pos = 0
    number = 0.0d0
    command = nop_

    do while(pos < lenfunc .and. .not. finish)
       pos = pos + 1

       select case(func(pos:pos))

       case ('0':'9')
          if (dflag) then
             exponent = .true.
             dflag = .false.
          end if

       case ('d','e','D','E')
          if (dflag .or. exponent) then
             finish = .true.
          else
             dflag = .true.
          end if

       case('.')
          if (dflag .or. exponent) finish = .true.

       case('+', '-')
          if (dflag) then
             dflag = .false.
             exponent = .true.
          else
             finish = .true.
          end if

       case default
          finish = .true.

       end select
    end do

    if (pos == lenfunc .and. .not. finish) then
       read(func, *) number

       command = number_

       func = ''
       lenfunc = 0
    else
       read(func(:pos-1), *) number

       command = number_

       func = func(pos:)
       lenfunc = lenfunc - pos + 1
    end if

  end subroutine s_readnumber



  subroutine s_readtext(func, lenfunc, command, number)
    integer, intent(inout) :: lenfunc
    character(lenfunc), intent(inout) :: func
    integer, intent(out) :: command
    real(fdp), intent(out) :: number
    integer :: i, funccut

    funccut = 0
    command = nop_
    number = 0.0d0

    do i = expressionstart, expressionend
       if (expressionslen(i) < lenfunc) then
          if ((func(:expressionslen(i)) == expressions(i)(:expressionslen(i))) &
               .and. expressionslen(i) > funccut ) then
             command = i
             funccut = expressionslen(i)-1
          end if
       end if
    end do

    if (command == nop_) then
       do i = 1, numvariables
          if (func(:min(lenvariables(i), lenfunc)) == variables(i) .and. &
               &lenvariables(i) > funccut) then
             command = variablestart+i
             number = varerrors(i)
             funccut = lenvariables(i)
          end if
       end do
    end if

    if (command == nop_) then
       call s_seterror('Unknown text: '//func)

       func = ''
       lenfunc = 0
    end if
    
    if (lenfunc > funccut) then
       func = func(funccut+1:)
       lenfunc = lenfunc - funccut
    else
       func = ''
       lenfunc = 0
    end if
  end subroutine s_readtext



  subroutine s_readoperator(func, lenfunc, command, order)
    integer, intent(inout) :: lenfunc
    character(lenfunc), intent(inout) :: func
    integer, intent(out) :: command, order
    integer :: i, funccut

    funccut = 0
    command = nop_
    order = huge(order)

    do i = operatorstart, operatorend
       if (func(:min(operatorslen(i), lenfunc)) == operators(i) .and. &
            &operatorslen(i) > funccut ) then
          command = i
          order = operatorsorder(i)
          funccut = operatorslen(i)
       end if
    end do

    if (command == nop_) then
       call s_seterror('Unknown operator: '//func)

       func = ''
       lenfunc = 0
    end if

    if (lenfunc > funccut) then
       func = func(funccut+1:)
       lenfunc = lenfunc - funccut
    else
       func = ''
       lenfunc = 0
    end if
  end subroutine s_readoperator



  subroutine s_readbracket(func, lenfunc, subfunc, lensubfunc)
    integer, intent(inout) :: lenfunc
    character(lenfunc), intent(inout) :: func
    character(len=len(func)), intent(out) :: subfunc
    integer, intent(out) :: lensubfunc
    integer :: pos, brackettype
    integer, dimension(bracketstart:bracketend) :: count

    count = 0
    brackettype = 0
    subfunc = ''
    
    select case(func(1:1))
       
    case('(')
       brackettype = roundbracket_
       
    case('[')
       brackettype = squarebracket_
       
    case('{')
       brackettype = squigbracket_
       
    end select

    do pos = 1, lenfunc
       select case(func(pos:pos))
          
       case('(')
          count(roundbracket_) = count(roundbracket_) + 1
          
       case(')')
          count(roundbracket_) = count(roundbracket_) - 1
          
       case('[')
          count(squarebracket_) = count(squarebracket_) + 1
          
       case(']')
          count(squarebracket_) = count(squarebracket_) - 1
          
       case('{')
          count(squigbracket_) = count(squigbracket_) + 1
          
       case('}')
          count(squigbracket_) = count(squigbracket_) - 1

       end select
       
       if (count(brackettype) == 0) then
          exit
       end if
    end do
    
    subfunc = func(2:pos-1)
    lensubfunc = pos - 2

    if (sum(abs(count)) /= 0) then
       call s_seterror('Unbalanced brackets: '//func)
       
       func = ''
       lenfunc = 0
    end if
    
    if (lensubfunc == 0) then
       call s_seterror('Empty brackets encountered: '//func)
       
       func = ''
       lenfunc = 0
    end if
    
    if (lenfunc < pos+1) then
       func = ''
       lenfunc = 0
    else
       func = func(pos+1:)
       lenfunc = lenfunc - pos
    end if
  end subroutine s_readbracket



  subroutine s_readcomma(func, lenfunc, command, order)
    integer, intent(inout) :: lenfunc
    character(lenfunc), intent(inout) :: func
    integer, intent(out) :: command, order
    
    command = comma_
    order = 0
    
    if (lenfunc < 2) then
       func = ''
       lenfunc = 0
    else
       func = func(2:)
       lenfunc = lenfunc - 1
    end if
  end subroutine s_readcomma



  recursive subroutine s_parseatom(func, lenfunc, funcstack, funcstackn, command, number, order, components)
    type(cachefn), intent(inout) :: funcstack
    integer, intent(inout) :: funcstackn
    integer, intent(inout) :: lenfunc, components
    character(lenfunc), intent(inout) :: func
    integer, intent(in) :: command, order
    real(fdp), intent(in) :: number
    character(len=len(func)) :: tmpfunc
    integer :: lentmpfunc, nextcommand, nextorder, newcomponents
    real(fdp) :: nextnumber

    select case(command)

    case(number_, variablestart:variableend)
       call s_funcpush(funcstack, funcstackn, command, number)

    case(nop_, end_)

    case(comma_)
       call s_parseblock(command, func, lenfunc, funcstack, funcstackn, order, components)
       
       components = components + 1
       
    case(bracket_)
       call s_readbracket(func, lenfunc, tmpfunc, lentmpfunc)
       call s_parsefn(tmpfunc, lentmpfunc, funcstack, funcstackn, components)

    case(operatorstart:operatorend)
       call s_parseblock(command, func, lenfunc, funcstack, funcstackn, order, components)
       call s_funcpush(funcstack, funcstackn, command, number)

    case(expressionstart:expressionend)
       call s_getnextatom(func, lenfunc, nextcommand, nextnumber, nextorder)
       call s_checkoperations(func, lenfunc, command, nextcommand)

       newcomponents = 1
       call s_parseatom(func, lenfunc, funcstack, funcstackn, nextcommand, nextnumber, nextorder, newcomponents)
       call s_checkcomponents(newcomponents, command)
       call s_funcpush(funcstack, funcstackn, command, number)
       
    case default
       call s_seterror('Unknown input: '//func)
       
       func = ''
       lenfunc = 0

    end select
  end subroutine s_parseatom



  subroutine s_getnextatom(func, lenfunc, command, number, order)
    integer, intent(inout) :: lenfunc
    character(lenfunc), intent(inout) :: func
    integer, intent(out) :: command, order
    real(fdp), intent(out) :: number

    command = nop_
    order = huge(order)
    number = 0.0d0

    if (lenfunc > 0) then
       select case(func(1:1))
          
       case('0':'9', '.')
          call s_readnumber(func, lenfunc, command, number)
          
       case('+', '-', '/', '*', '^', '>', '!', '<', '=')
          call s_readoperator(func, lenfunc, command, order)
          
       case('a':'z','A':'Z')
          call s_readtext(func, lenfunc, command, number)
          
       case('(','[','{')
          command = bracket_

       case(',')
          call s_readcomma(func, lenfunc, command, order)
          
       case default
          if (lenfunc > 1) call s_seterror('Unknown input: '//func)
          
          func = ''
          lenfunc = 0
          command = end_
          
       end select
    else
       command = end_
    end if
  end subroutine s_getnextatom
  


  subroutine s_peeknextatom(func, lenfunc, command, number, order)
    integer, intent(in) :: lenfunc
    character(lenfunc), intent(in) :: func
    integer, intent(out) :: command, order
    real(fdp), intent(out) :: number
    character(len=lenfunc) :: tmpfunc
    integer :: lentmpfunc

    command = nop_
    order = huge(order)
    tmpfunc = func
    lentmpfunc = lenfunc

    call s_getnextatom(tmpfunc, lentmpfunc, command, number, order)
  end subroutine s_peeknextatom



  recursive subroutine s_parseblock(command, func, lenfunc, funcstack, n, blockorder, components)
    integer, intent(inout) :: lenfunc
    character(lenfunc), intent(inout) :: func
    type(cachefn), intent(inout) :: funcstack
    integer, intent(inout) :: n, components
    integer, intent(in) :: command, blockorder
    integer :: tcommand, nextcommand, nextorder
    real(fdp) :: nextnumber

    tcommand = command
    nextcommand = nop_
    nextorder = huge(nextorder)

    call s_peeknextatom(func, lenfunc, nextcommand, nextnumber, nextorder)
    call s_checkoperations(func, lenfunc, tcommand, nextcommand)

    if ((nextcommand == plus_ .or. nextcommand == minus_) .and. &
         count(command == blockseparator) > 0) call s_funcpush(funcstack, n, number_, 0.0d0)

    do while(nextorder > blockorder .and. nextcommand /= end_)
       call s_getnextatom(func, lenfunc, nextcommand, nextnumber, nextorder)
       call s_parseatom(func, lenfunc, funcstack, n, nextcommand, nextnumber, nextorder, components)
       tcommand = nextcommand
       call s_peeknextatom(func, lenfunc, nextcommand, nextnumber, nextorder)
    end do
  end subroutine s_parseblock



  recursive subroutine s_parsefn(func, lenfunc, funcstack, funcstackn, components)
    integer, intent(inout) :: lenfunc
    character(lenfunc), intent(inout) :: func
    type(cachefn), intent(inout) :: funcstack
    integer, intent(inout) :: funcstackn, components
    integer :: order, command
    real(fdp) :: number

    call s_getnextatom(func, lenfunc, command, number, order)
    
    if (command == plus_ .or. command == minus_) call s_funcpush(funcstack, funcstackn, number_, 0.0d0)
    
    do while(command /= end_)
       call s_parseatom(func, lenfunc, funcstack, funcstackn, command, number, order, components)
       call s_getnextatom(func, lenfunc, command, number, order)
    end do
  end subroutine s_parsefn


  subroutine s_evaluate(funcstack, funcstackn, varvals, result, errvals)
    type(cachefn), intent(in) :: funcstack
    integer, intent(inout) :: funcstackn
    real(fdp), dimension(:), intent(in) :: varvals
    real(fdp), dimension(:), intent(inout), optional :: errvals
    real(fdp), intent(out) :: result
    real(fdp), dimension(:), allocatable :: tmperrvals
    real(fdp), dimension(:), save, allocatable :: stack
    real(fdp), dimension(maxcommandcomps) :: tmp
    integer :: command, stackn
    real(fdp) :: number

    if (.not. allocated(stack)) then
       allocate(stack(0:size(funcstack%numbers)))
       stack = 0.0d0
    elseif (size(stack) < size(funcstack%numbers)+1) then
       deallocate(stack)
       allocate(stack(0:size(funcstack%numbers)))
       stack = 0.0d0
    end if

    command = nop_
    number = 0.0d0
    result = 0.0d0
    stackn = 0

    if (present(errvals)) then
       allocate(tmperrvals(size(errvals)))
       tmperrvals = errvals
    end if

    do while(command /= end_)
       funcstackn=funcstackn+1;command=funcstack%commands(funcstackn);number=funcstack%numbers(funcstackn)

       select case(command)

       case(number_)
          stackn=stackn+1;stack(stackn)=result
          result = number

       case(variablestart:variableend)
          stackn=stackn+1;stack(stackn)=result
          result = varvals(command-variablestart)
          if (present(errvals)) then
             tmperrvals(command-variablestart) = errvals(command-variablestart)*number
             result = result+tmperrvals(command-variablestart)
          end if
             
       case(pow_, caret_)
          tmp(1)=stack(stackn);stackn=stackn-1

          if (tmp(1) > 0.0d0) then
             result = tmp(1) ** result
          elseif (tmp(1) < 0.0d0) then
             if (aint(result) == result) then
                result = tmp(1) ** nint(result)
             else
                call s_seterror('x**y, x<0 and y not integer')
             end if
          elseif (result /= 0.0d0) then
             result = 0.0d0
          else             
             call s_seterror('Undefined result 0**0')
          end if

       case(neq_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = merge(1.0d0, 0.0d0, tmp(1) /= result)

       case(eqeq_, eq_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = merge(1.0d0, 0.0d0, tmp(1) == result)

       case(gteq_, eqgt_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = merge(1.0d0, 0.0d0, tmp(1) >= result) 

       case(lteq_, eqlt_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = merge(1.0d0, 0.0d0, tmp(1) <= result)

       case(gt_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = merge(1.0d0, 0.0d0, tmp(1) > result)

       case(lt_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = merge(1.0d0, 0.0d0, tmp(1) < result)

       case(mult_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = tmp(1) * result

       case(div_)
          tmp(1)=stack(stackn);stackn=stackn-1

          if (result /= 0.0d0) then
             result = tmp(1) / result
          else
             call s_seterror('x/y, y=0')
          end if

       case(plus_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = tmp(1) + result

       case(minus_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = tmp(1) - result

       case(cos_)
          result = cos(result)

       case(sin_)
          result = sin(result)

       case(tan_)
          result = tan(result)

       case(exp_)
          result = exp(result)

       case(log_)
          if (result > 0.0d0) then
             result = log(result)
          else
             call s_seterror('log(x), x<=0')
          end if

       case(log10_)
          if (result > 0.0d0) then
             result = log10(result)
          else
             call s_seterror('log10(x), x<=0')
          end if

       case(sqrt_)
          if (result >= 0.0d0) then
             result = sqrt(result)
          else
             call s_seterror('sqrt(x), x<0')
          end if

       case(cbrt_)
          result = f_cbrt(result)

       case(acos_)
          if (abs(result) <= 1.0d0) then
             result = acos(result)
          else
             call s_seterror('acos(x), |x|>1')
          end if

       case(asin_)
          if (abs(result) <= 1.0d0) then
             result = asin(result)
          else
             call s_seterror('asin(x), |x|>1')
          end if

       case(atan_)
          result = atan(result)

       case(cosh_)
          result = cosh(result)

       case(sinh_)
          result = sinh(result)

       case(tanh_)
          result = tanh(result)

       case(anint_)
          result = anint(result)

       case(aint_)
          result = aint(result)

       case(abs_)
          result = abs(result)

       case(delta_)
          result = merge(1.0d0, 0.0d0, result == 0.0d0)

       case(step_)
          result = merge(0.0d0, 1.0d0, result < 0.0d0)

       case(hat_)
          result = merge(1.0d0, 0.0d0, result <= 0.5d0 .and. result >= -0.5d0)

       case(max_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = max(result, tmp(1))

       case(min_)
          tmp(1)=stack(stackn);stackn=stackn-1
          result = min(result, tmp(1))

       case(besj0_)
          result = f_besj0(result)

       case(besj1_)
          result = f_besj1(result)

       case(besjn_)
          tmp(1)=stack(stackn);stackn=stackn-1
          if (aint(tmp(1)) == tmp(1)) then
             result = f_besjn(tmp(1), result)
          else
             call s_seterror('besjn(n, x), n not integer')
          end if

       case(besy0_)
          result = f_besy0(result)

       case(besy1_)
          result = f_besy1(result)

       case(besyn_)
          tmp(1)=stack(stackn);stackn=stackn-1
          if (aint(tmp(1)) == tmp(1) .and. result > 0.0d0) then
             result = f_besyn(tmp(1), result)
          else
             call s_seterror('besyn(n, x), n not integer or x/>0')
          end if
          
       case(logn_)
          tmp(1)=stack(stackn);stackn=stackn-1

          if (result > 0.0d0 .and. tmp(1) > 0.0d0) then
             result = log(result)/log(tmp(1))
          else
             call s_seterror('logn(x, y), x/>0 or y/>0')
          end if

       case(erf_)
          result = f_erf(result)

       case(erfc_)
          result = 1.0d0 - f_erf(result)

       case(lgamma_)
          if (result > 0.0d0) then
             result = f_lgamma(result)
          else
             call s_seterror('lgamma(x), x<=0')
          end if

       case(gamma_)
          result = f_gamma(result)

       case(csch_)
          result = sinh(result)

          if (result == 0.0d0) then
             call s_seterror('1/sinh(x), sinh(x) = 0')
          else
             result = 1.0d0/result
          end if

       case(sech_)
          result = 1.0d0/cosh(result)

       case(coth_)
          result = tanh(result)

          if (result == 0.0d0) then
             call s_seterror('1/tanh(x), tanh(x) = 0')
          else
             result = 1.0d0/result
          end if

       case(if_)
          tmp(1)=stack(stackn);stackn=stackn-1
          tmp(2)=stack(stackn);stackn=stackn-1

          result = merge(result, tmp(1), tmp(2) == 0)

       case(gauss_)
          result = exp(-result**2)

       case(sinc_)
          result = merge(1.0d0, sin(result)/result, result == 0.0d0)

       case(besi0_)
          result = f_besi0(result)

       case(besi1_)
          result = f_besi1(result)
          
       case(besin_)
          tmp(1)=stack(stackn);stackn=stackn-1
          if (aint(tmp(1)) == tmp(1)) then
             result = f_besin(tmp(1), result)
          else
             call s_seterror('besin(n, x), n not integer')
          end if

       case(besk0_)
          result = f_besk0(result)
          
       case(besk1_)
          result = f_besk1(result)
          
       case(beskn_)
          tmp(1)=stack(stackn);stackn=stackn-1
          if (aint(tmp(1)) == tmp(1)) then
             result = f_beskn(tmp(1), result)
          else
             call s_seterror('besjk(n, x), n not integer')
          end if

       case(ierfc_)
          if (abs(result-1.0d0) < 1.0d0) then
             result = f_ierfc(result)
          else
             call s_seterror('ierfc(x), 0 < x < 2')
          end if

       case(fresc_)
          result = f_fresc(result)

       case(fress_)
          result = f_fress(result)

       case(expi_)
          result = f_expi(result)

       case(sini_)
          result = f_sini(result)

       case(cosi_)
          result = f_cosi(result)

       case(logi_)
          if (result > 0.0d0) then
             result = f_logi(result)
          else
             call s_seterror('logi(x), x > 0')
          end if

       case(ierf_)
          if (abs(result) < 1.0d0) then
             result = f_ierf(result)
          else
             call s_seterror('ierf(x), -1 < x < 1')
          end if

       case(elle_)
          if (abs(result) < 1.0d0) then
             result = f_elle(result)
          else
             call s_seterror('elle(x), -1 < x < 1')
          end if

       case(ellk_)
          if (abs(result) < 1.0d0) then
             result = f_ellk(result)
          else
             call s_seterror('ellk(x), -1 < x < 1')
          end if

       case(ielle_)
          tmp(1)=stack(stackn);stackn=stackn-1
          if (abs(tmp(1)) < 1.0d0 .and. abs(result) < pi2) then
             result = f_ielle(tmp(1), result)
          else
             call s_seterror('ielle(x, phi), -1 < x < 1, -pi/2 < phi < pi/2')
          end if

       case(iellf_)
          tmp(1)=stack(stackn);stackn=stackn-1
          if (abs(tmp(1)) < 1.0d0 .and. abs(result) < pi2) then
             result = f_iellf(tmp(1), result)
          else
             call s_seterror('iellf(x, phi), -1 < x < 1, -pi/2 < phi < pi/2')
          end if

       end select
    end do

    if (stackn > 1) call s_seterror('Evaluation stack not empty')

    if (present(errvals)) then
       errvals = tmperrvals
       deallocate(tmperrvals)
    end if
  end subroutine s_evaluate



  subroutine s_variableerr(funcstack, number, varvals, errval, i)
    type(cachefn), intent(in) :: funcstack
    real(fdp), intent(inout) :: number
    real(fdp), dimension(:), intent(in) :: varvals
    real(fdp), intent(in) :: errval
    integer, intent(in) :: i
    real(fdp), dimension(size(varvals)) :: errvals
    real(fdp) :: numplus, numminus, rangeplus, rangeminus
    integer :: funcstackn

    number = 0.0d0

    if (error == 'OK' .and. size(varvals) > 0) then
       !
       ! Compute actual error
       !

       errvals = 0.0d0
       errvals(i) = errval
       funcstackn = 0

       call s_evaluate(funcstack, funcstackn, varvals, numplus, errvals)

       rangeplus = errvals(i)
       errvals(i) = -errval
       funcstackn = 0

       call s_evaluate(funcstack, funcstackn, varvals, numminus, errvals)

       rangeminus = errvals(i)

       number = merge(0.5d0*(numplus*rangeminus**2-numminus*rangeplus**2-&
            number*(rangeminus+rangeplus)**2)/rangeplus/rangeminus, &
            0.0d0, (rangeminus /= 0) .and. (rangeplus /= 0))
    end if
  end subroutine s_variableerr
  


  subroutine s_createfn(func, vars, funcnumber, t_error)
    character(len=*), intent(in) :: func, vars
    integer, intent(out) :: funcnumber
    character(len=*), intent(out), optional :: t_error
    type(cachefn), pointer :: funcstack
    type(cachefn), dimension(:), pointer :: tmpcachedfuncs
    integer :: funcstackn, i, j, n, lentmpfunc, components
    character(len=len(func)) :: tmpfunc
    
    funcnumber = 0
    error = 'OK'

    call s_firstrun()

    if (trim(vars) /= "") call s_allocatevariables(vars)
    
    if (error == 'OK') then
       n = size(cachedfuncs)

       do i=1, n
          if (.not. associated(cachedfuncs(i)%numbers)) exit
       end do

       if (i > n) then
          allocate(tmpcachedfuncs(i+10))

          do j = 1, n
             tmpcachedfuncs(j)%commands => cachedfuncs(j)%commands
             tmpcachedfuncs(j)%numbers => cachedfuncs(j)%numbers
          end do
          
          do j = n+1, size(tmpcachedfuncs)
             nullify(tmpcachedfuncs(j)%commands, tmpcachedfuncs(j)%numbers)
          end do

          deallocate(cachedfuncs)
          cachedfuncs => tmpcachedfuncs
       end if

       funcnumber = i
       funcstack => cachedfuncs(i)
       funcstackn = 0
       tmpfunc = f_clean(func)
       lentmpfunc = len_trim(tmpfunc)
       
       allocate(funcstack%commands(10), funcstack%numbers(10))
       
       funcstack%commands = end_
       funcstack%numbers = 0.0d0
       funcstackn = 0
       components = 0

       call s_parsefn(tmpfunc, lentmpfunc, funcstack, funcstackn, components)

!       if (components > 1) call s_seterror('Multiple component function')
    end if

    if (trim(vars) /= "") call s_deallocatevariables()
    
    if (present(t_error)) t_error = error
  end subroutine s_createfn



  subroutine s_destroyfn(funcnumber)
    integer, intent(in) :: funcnumber

    call s_firstrun()
    
    if (funcnumber > size(cachedfuncs)) then
       call s_seterror('Function does not exist')
    else if (associated(cachedfuncs(funcnumber)%commands)) then
       deallocate(cachedfuncs(funcnumber)%numbers, cachedfuncs(funcnumber)%commands)
    else
       call s_seterror('Function does not exist')
    end if
  end subroutine s_destroyfn
  
  
  
  subroutine s_evaluatefn(funcnumber, varvals, number, t_error)
    integer, intent(in) :: funcnumber
    real(fdp), dimension(:), intent(in) :: varvals
    real(fdp), intent(out) :: number
    character(len=*), intent(out), optional :: t_error
    type(cachefn) :: funcstack
    integer :: funcstackn

    error = 'OK'
    
    call s_firstrun()

    number = 0.0d0
    if (present(t_error)) t_error = ''
    
    if (funcnumber > size(cachedfuncs) .or. funcnumber < 1) then
       call s_seterror('Function does not exist')
    else if (associated(cachedfuncs(funcnumber)%commands)) then
       funcstack = cachedfuncs(funcnumber)
       funcstackn = 0

       call s_evaluate(funcstack, funcstackn, varvals, number)
    else
       call s_seterror('Function does not exist')
    end if

    if (present(t_error)) t_error = error
  end subroutine s_evaluatefn
  
  
  
  subroutine s_evaluateerr(funcnumber, varvals, number, t_error)
    integer, intent(in) :: funcnumber
    real(fdp), dimension(:), intent(in) :: varvals
    real(fdp), intent(out) :: number
    character(len=*), intent(out), optional :: t_error
    type(cachefn) :: funcstack
    real(fdp), dimension(size(varvals)) :: errvals
    real(fdp) :: num, halfnumber, tmp
    integer :: funcstackn, i

    error = 'OK'

    call s_firstrun()

    number = 0.0d0
    if (present(t_error)) t_error = ''
    halfnumber = 0.0d0

    if (funcnumber > size(cachedfuncs) .or. funcnumber < 1) then
       call s_seterror('Function does not exist')
    else if (associated(cachedfuncs(funcnumber)%commands)) then
       funcstack = cachedfuncs(funcnumber)
       errvals = 0.0d0
       funcstackn = 0
       
       call s_evaluate(funcstack, funcstackn, varvals, num, errvals)

       do i = 1, size(varvals)
          tmp = num

          call s_variableerr(funcstack, tmp, varvals, 1.0d0, i)

          number = number + abs(tmp)

          !
          ! Compute the error with half the range
          !

          tmp = num

          call s_variableerr(funcstack, tmp, varvals, 0.5d0, i)

          halfnumber = halfnumber + abs(tmp)
       end do

       !
       ! Return correct precision error
       !

       halfnumber = 2.0d0*halfnumber

       call s_errprecision(number)
       call s_errprecision(halfnumber)

       !
       ! Verify error value is consistent
       !

       if (halfnumber /= number) call s_seterror('Error value not stable')
    else
       call s_seterror('Function does not exist')
    end if
    
    if (present(t_error)) t_error = error
  end subroutine s_evaluateerr
  
  
  
end module interpreter